(***********************************************************************)
(*                                                                     *)
(*                           Objective Caml                            *)
(*                                                                     *)
(*            Xavier Leroy, projet Cristal, INRIA Rocquencourt         *)
(*                                                                     *)
(*  Copyright 1996 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: linearize.ml,v 1.2 1997/09/09 17:47:59 blanchet Exp $ *)

(* Transformation of Mach code into a list of pseudo-instructions. *)

open Reg
open Mach

type label = int

let label_counter = ref 99

let new_label() = incr label_counter; !label_counter

type t_stack_objects = (int * int) list

type instruction =
  { mutable desc: instruction_desc;
    next: instruction;
    arg: Reg.t array;
    res: Reg.t array;
    stack_alloc: int;               (* current state of the stack *)
    stack_objects: t_stack_objects;
    live: Reg.Set.t }

and instruction_desc =
    Lend
  | Lop of operation
  | Lreloadretaddr
  | Lreturn
  | Llabel of label
  | Lbranch of label
  | Lcondbranch of test * label
  | Lcondbranch3 of label option * label option * label option
  | Lswitch of label array
  | Lsetuptrap of label
  | Lpushtrap
  | Lpoptrap
  | Lraise

type fundecl =
  { fun_name: string;
    fun_body: instruction;
    fun_fast: bool }

(* Invert a test *)

let invert_integer_test = function
    Isigned cmp -> Isigned(Cmm.negate_comparison cmp)
  | Iunsigned cmp -> Iunsigned(Cmm.negate_comparison cmp)

let invert_test = function
    Itruetest -> Ifalsetest
  | Ifalsetest -> Itruetest
  | Iinttest(cmp) -> Iinttest(invert_integer_test cmp)
  | Iinttest_imm(cmp, n) -> Iinttest_imm(invert_integer_test cmp, n)
  | Ifloattest(cmp, neg) -> Ifloattest(cmp, not neg)
  | Ieventest -> Ioddtest
  | Ioddtest -> Ieventest

(* The "end" instruction *)

let rec end_instr =
  { desc = Lend;
    next = end_instr;
    arg = [||];
    res = [||];
    stack_alloc = 0;
    stack_objects = [];
    live = Reg.Set.empty }

(* Cons an instruction (live empty) *)

let instr_cons d a r sa so n =
  { desc = d; next = n; arg = a; res = r; stack_alloc = sa; stack_objects =
    so;  live = Reg.Set.empty }

(* Cons a simple instruction (arg, res, live empty) *)

let cons_instr d sa so n =
  { desc = d; next = n; arg = [||]; res = [||]; stack_alloc = sa;
    stack_objects = so; live = Reg.Set.empty }

(* Build an instruction with arg, res, live taken from
   the given Mach.instruction *)

let copy_instr d i sa so n =
  { desc = d; next = n;
    arg = i.Mach.arg; res = i.Mach.res; 
    stack_alloc = sa; stack_objects = so; live = i.Mach.live }

(* Label the beginning of the given instruction sequence.
   If the sequence starts with a branch, jump over it. *)

let get_label sa so n =
  match n.desc with
    Lbranch lbl -> (lbl, n)
  | Llabel lbl -> (lbl, n)
  | _ -> let lbl = new_label() in (lbl, cons_instr (Llabel lbl) sa so n)

(* Discard all instructions up to the next label.
   This function is to be called before adding a non-terminating 
   instruction. *)

let rec discard_dead_code n =
  match n.desc with
    Lend -> n
  | Llabel _ -> n
(* Do not discard Lpoptrap or Istackoffset instructions,
   as this may cause a stack imbalance later during assembler generation.
   Not true any more - Blanchet  
  | Lpoptrap -> n
  | Lop(Istackoffset _) -> n *)
  | _ -> discard_dead_code n.next

(* Add a branch in front of a continuation.
   Discard dead code in the continuation.
   Does not insert anything if we're just falling through. *)

let add_branch lbl sa so n =
  let n1 = discard_dead_code n in
  match n1.desc with
    Llabel lbl1 when lbl1 = lbl -> n1
  | _ -> cons_instr (Lbranch lbl) sa so n1

(* Current label for exit handler *)

let exit_label = ref 99

(* Record stack-allocated objects *)

let record_stack_object size sa so =
   (sa, size) :: so

let free_stack_objects sa so =
    let rec free = function
     [] -> []
   | (depth, size) :: rem as objlist ->
       if depth > sa then free rem else objlist in
   free so

(* Linearize an instruction [i]: add it in front of the continuation [n] *)

let rec linear i sa so n =
  match i.Mach.desc with
    Iend -> n
  | Iop(Itailcall_ind | Itailcall_imm _ as op) ->
      copy_instr (Lop op) i sa so (discard_dead_code n)
  | Iop(Imove | Ireload | Ispill)
    when i.Mach.arg.(0).loc = i.Mach.res.(0).loc ->
      linear i.Mach.next sa so n
  | Iop op ->
      begin
      match op with
         Istackoffset sz ->
           begin
             match i.Mach.next.Mach.desc with
                Ireturn -> linear i.Mach.next sa so n
                    (* Ireturn flushes the whole stack anyway *)
              | _ ->
                  let sz = Misc.align sz Proc.stack_align in
                  let n' = linear i.Mach.next (sa+sz)
                     (if sz<0 then free_stack_objects (sa+sz) so else so) n
in
                    match n'.desc with
                       Lop(Istackoffset sz') -> if sz+sz' <> 0 then
                               copy_instr (Lop (Istackoffset
                                  (sz+sz'))) i.Mach.next sa so n'.next
                            else n'.next
                     | _ -> copy_instr (Lop (Istackoffset sz)) i sa so n'
           end
       | Istackalloc n0 ->
             let na = Misc.align n0 Proc.stack_align in
             let sa' = sa + na
             and so' = record_stack_object n0 (sa+na) so in
             cons_instr (Lop(Istackoffset na)) sa so
             (copy_instr (Lop (Istackalloc 0)) i sa' so'
             (linear i.Mach.next sa' so' n))
       | Ispecific _ -> copy_instr (Lop op) i sa so (linear
             i.Mach.next (sa + Proc.specific_stack_offset i) so n)
       | _ -> copy_instr (Lop op) i sa so (linear i.Mach.next sa so n)
      end
  | Ireturn ->
      let n1 = copy_instr Lreturn i sa so (discard_dead_code n) in
      if !Proc.contains_calls
      then cons_instr Lreloadretaddr sa so n1
      else n1
  | Iifthenelse(test, ifso, ifnot) ->
      let n1 = linear i.Mach.next sa so n in
          (* because of good parenthesing of stack allocations *)
       begin match (ifso.Mach.desc, ifnot.Mach.desc) with
        Iexit, _ ->
          copy_instr (Lcondbranch(test, !exit_label)) i sa so
            (linear ifnot sa so n1)
      | _,  Iexit ->
          copy_instr (Lcondbranch(invert_test test, !exit_label)) i sa so
            (linear ifso sa so n1)
      | Iend, _ ->
          let (lbl_end, n2) = get_label sa so n1 in
          copy_instr (Lcondbranch(test, lbl_end)) i sa so
            (linear ifnot sa so n2)
      | _,  Iend ->
          let (lbl_end, n2) = get_label sa so n1 in
          copy_instr (Lcondbranch(invert_test test, lbl_end)) i sa so
            (linear ifso sa so n2)
      | _, _ ->
        (* Should attempt branch prediction here *)
          let (lbl_end, n2) = get_label sa so n1 in
          let (lbl_else, nelse) = get_label sa so (linear ifnot sa so n2) in
          copy_instr (Lcondbranch(invert_test test, lbl_else)) i sa so 
            (linear ifso sa so (add_branch lbl_end sa so nelse))
      end
  | Iswitch(index, cases) ->
      let lbl_cases = Array.create (Array.length cases) 0 in
      let (lbl_end, n1) = get_label sa so (linear i.Mach.next sa so n) in
      let n2 = ref (discard_dead_code n1) in
      for i = Array.length cases - 1 downto 0 do
        let (lbl_case, ncase) =
                get_label sa so (linear cases.(i) sa so (add_branch lbl_end
sa so !n2)) in
        lbl_cases.(i) <- lbl_case;
        n2 := discard_dead_code ncase
      done;
      (* Switches with 1 and 2 branches have been eliminated earlier.
         Here, we do something for switches with 3 branches. *)
      if Array.length index = 3 then begin
        let fallthrough_lbl =
          match !n2.desc with Llabel lbl -> lbl | _ -> -1 in
        let find_label n =
          let lbl = lbl_cases.(index.(n)) in
          if lbl = fallthrough_lbl then None else Some lbl in
        copy_instr (Lcondbranch3(find_label 0, find_label 1, find_label 2))
                   i sa so !n2
      end else
        copy_instr (Lswitch(Array.map (fun n -> lbl_cases.(n)) index)) i sa
so !n2
  | Iloop body ->
      let lbl_head = new_label() in
      let n1 = linear i.Mach.next sa so n in
      let n2 = linear body sa so (cons_instr (Lbranch lbl_head) sa so n1) in
      cons_instr (Llabel lbl_head) sa so n2
  | Icatch(body, handler) ->
      let (lbl_end, n1) = get_label sa so (linear i.Mach.next sa so n) in
      let (lbl_handler, n2) = get_label sa so (linear handler sa so n1) in
      let saved_exit_label = !exit_label in
      exit_label := lbl_handler;
      let n3 = linear body sa so (add_branch lbl_end sa so n2) in
      exit_label := saved_exit_label;
      n3
  | Iexit ->
      add_branch !exit_label sa so (linear i.Mach.next sa so n)
  | Itrywith(body, handler) ->
      let sa' = sa + Proc.trap_size in
      let (lbl_join, n1) = get_label sa so (linear i.Mach.next sa so n) in
      let (lbl_body, n2) =
        get_label sa so (cons_instr Lpushtrap sa so 
                    (linear body sa' so (cons_instr Lpoptrap sa' so n1))) in
      cons_instr (Lsetuptrap lbl_body) sa so 
        (linear handler sa so (add_branch lbl_join sa so n2))
  | Iraise ->
      copy_instr Lraise i sa so (discard_dead_code n)

let fundecl f =
  { fun_name = f.Mach.fun_name;
    fun_body = linear f.Mach.fun_body 0 [] end_instr;
    fun_fast = f.Mach.fun_fast }

