(***********************************************************************)
(*                                                                     *)
(*              Fast Escape Analysis - Equations solver                *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                  Alain Deutsch, INRIA Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Typedtree
open Format

let top = 32000

module type QueueType =
  sig
    type data

    type t = {mutable key : int;
              mutable index : int;
              mutable l_eqlist : data;
              mutable r_eqlist : data; 
              name : string}
  end


module PriorityQueue(Data : QueueType) =
struct

(* Priority queue implemented with binary heaps *)

     exception EmptyQueue

     type t = Data.t;;

     open Data

     let rec update (taille,q) node key = let i = node.index in
        if node.key < key then
            let pere = (i-1)/2 in
               if (i>0) && (q.(pere).key < key) then
               begin
                  q.(i) <- q.(pere);
                  q.(pere) <- node;
                  node.index <- pere;
                  q.(i).index <- i;
                  update (taille,q) node key
               end
               else
                  node.key <- key

     let create l = 
        let taille = List.length l in
        let q = Array.create taille (List.hd l) in
        let queue = (ref taille,q) in
        let i = ref 0 in
        let listkey = List.map (fun elem -> q.(!i) <- elem; 
            let key = elem.key in
                elem.key <- 0;
                elem.index <- !i;
                incr i;
                key) l in
            List.iter2 (fun key elem -> update queue elem key) listkey l;
            queue;;

     let rec entasser (taille,q) i =
         let l = 2*i+1 in
         let max1 = if (l < !taille) && (q.(l).key > q.(i).key) then l else i in
         let r = 2*i+2 in
         let max2 = if (r < !taille) && (q.(r).key > q.(max1).key) then r else max1 in
            if max2 <> i then
            begin
               let tmp = q.(max2) in
                  q.(max2) <- q.(i);
                  q.(i) <- tmp;
                  q.(max2).index <- max2;
                  q.(i).index <- i;
                  entasser (taille,q) max2
            end
              

     let extractmax (taille,q) = if !taille = 0 then
           raise EmptyQueue
        else
           let max = q.(0) in
              q.(0) <- q.(!taille-1);
              q.(0).index <- 0;
              decr taille;
              entasser (taille,q) 0;
              max

end



module KnuthEquations =
struct

type term = Term of int
          | Nonterm of nonterm
          | Min of nonterm * nonterm
          | Minc of int * nonterm
          | NonKnuthFun of nonterm * (int -> int) * int
          | KnuthFun of nonterm * (int -> int)

and equation = {lhs : nonterm; mutable rhs :  term; mutable count : int}

and data = equation list

and t = {mutable key : int;
         mutable index : int;
         mutable l_eqlist : data;
         mutable r_eqlist : data;
         name : string}

and nonterm = t;;

(* Current value * equations having this non terminal as lhs * 
   equations using this non terminal on their rhs * index * name

   index is 0 when building the equations, >=0 when solving them,
   and -1 when equations are solved, ie when the current value won't change
   any more *)

end

module KnuthPriorityQueue = PriorityQueue(KnuthEquations)

module KnuthSolve :
sig 

open KnuthEquations
val reset : unit -> unit
    (* Resets the system *)
val newnonterm : string -> nonterm
    (* [newnonterm name] creates a new non terminal *)
val compute : term -> int
val addeq : nonterm -> term -> unit
    (* [addeq lhs rhs] adds an equation *)
val termmin : term -> term -> term
val termmax : term -> term -> term
val nonknuthfun : term -> (int -> int) -> int -> term
    (* [nonknuthfun t f kf] creates a term whose value is f(t) except
       if it is in a cycle, where we cannot use non Knuth functions,
       and where we take the constant approximation kf. We must have
       kf >= max { f(t) | t term } for correctness *)
val knuthfun : term -> (int -> int) -> term
    (* [knuthfun t f] creates a term whose value is f(t).
       We must have f(t) <= t to be able to apply the Knuth algorithm *)
val cfc : unit -> unit
    (* Solves the system *)
end =
struct

open KnuthEquations

(* two phases :

1. maintain a list of non-terminals with a list of the equations which use it 
   on their right hand-side, and the current key

2. solve the system by the Knuth algorihtm, after building a priority queue
   from the list mentionned above
*)

    let syst = ref []

    let reset() = (syst := [])

    let solved = -1
    let black = -2
    let gray = -3
    let white = -4

    let newnonterm name = let nonterm = {key = 0; l_eqlist = []; r_eqlist = []; index = 0; name = name} in
          syst:= nonterm:: !syst;
          nonterm

    let simplifterm nonterm_effect keep_inst t = match t with
              Term _ -> t
            | Nonterm nonterm -> if nonterm.index = solved then
                    Term nonterm.key
                 else
                 begin
                    nonterm_effect nonterm;
                    t
                 end
            | Minc(i1,t2) -> if t2.index = solved then
                    Term (min i1 t2.key)
                 else
                 begin
                    nonterm_effect t2;
                    t
                 end
            | Min(t1,t2) -> 
                 if t1.index<>solved then nonterm_effect t1;
                 if t2.index<>solved then nonterm_effect t2;
                 if (t1.index = solved) & (t2.index = solved) then
                   Term (min t1.key t2.key)
                 else t
            | NonKnuthFun(t1,f,kf) -> 
                 if t1.index = solved then
                    Term (f t1.key)
                 else if keep_inst then (nonterm_effect t1; t) else Term kf
            | KnuthFun(t1,f) -> if t1.index = solved then
                    Term (f t1.key)
                 else
                 begin
                    nonterm_effect t1;
                    t
                 end

    let simplif0 keep_inst t = match t with
              Term _ -> t
            | Nonterm nonterm -> if nonterm.index = solved then
                    Term nonterm.key
                 else
                    t
            | Minc(i1,t2) -> if t2.index = solved then
                    Term (min i1 t2.key)
                 else
                    t
            | Min(t1,t2) -> 
                 if (t1.index = solved) & (t2.index = solved) then
                   Term (min t1.key t2.key)
                 else t
            | NonKnuthFun(t1,f,kf) -> 
                 if t1.index = solved then
                    Term (f t1.key)
                 else if keep_inst then t else Term kf
            | KnuthFun(t1,f) -> if t1.index = solved then
                    Term (f t1.key)
                 else
                    t

    let compute t = 
       match simplif0 false t with
         Term n -> n
       | _ -> Message.debug 0
	(fun () -> print_string "Non terminal value used when computation not finished !!"); 0

    let addeq lhs rhs = 
       let eq = {lhs = lhs; rhs =rhs; count=0} in
       lhs.l_eqlist <- eq :: lhs.l_eqlist;
       eq.rhs <- simplifterm (fun nonterm -> 
          nonterm.r_eqlist <- eq :: nonterm.r_eqlist) true rhs;
       match eq.rhs with
               Term res1 -> lhs.key <- max lhs.key res1 
             | _ -> ()

    let tononterm = function Nonterm t -> t
                        | t -> let t' = newnonterm "tont" in
                             addeq t' t;
                             t'

    let termmin t1 t2 = match (simplif0 true t1,simplif0 true t2) with
           (_,Term 0) | (Term 0,_) -> Term 0
         | (Term i1, Term i2) -> Term(min i1 i2)
         | (Term i1,x2) -> if i1 = top then x2 else Minc(i1,tononterm x2)
         | (x1,Term i2) -> if i2 = top then x1 else Minc(i2,tononterm x1)
         | (x1,x2) -> Min(tononterm x1,tononterm x2)

    let termmax t1 t2 = match (simplif0 true t1,simplif0 true t2) with
           (Term i1, Term i2) -> Term(max i1 i2)
         | (Term 0, x2) -> x2
         | (x1, Term 0) -> x1
         | (x1,x2) -> let x' = newnonterm "maxinterm" in
              addeq x' x1;
              addeq x' x2;
              Nonterm x'

    let nonknuthfun t f kf = match (simplif0 true t) with
         Term n -> Term (f n)
       | _ -> NonKnuthFun(tononterm t, f, kf)

    let knuthfun t f = match (simplif0 true t) with
         Term n -> Term (f n)
       | _ -> KnuthFun(tononterm t, f)

    let knuth_addeq eq =
       eq.rhs <- simplifterm (fun nonterm -> 
                    nonterm.r_eqlist <- eq :: nonterm.r_eqlist;
                    eq.count<-eq.count+1) false eq.rhs;
       match eq.rhs with
               Term res1 ->  eq.lhs.key <- max eq.lhs.key res1 
             | _ -> ()

    let knuth_reset k =
       List.iter (fun nonterm -> nonterm.r_eqlist <- []) k;
       List.iter (fun nonterm ->
           List.iter knuth_addeq nonterm.l_eqlist) k

    let print_nonterm {name = name; index = index} =
       print_string name;
       print_int index

    let print_term = function
           Term n -> print_int n
         | Nonterm nonterm -> 
              if nonterm.index <= -1 then print_int nonterm.key else
                 print_nonterm nonterm
         | Minc(i1,t2) ->
              print_string "min(";
              print_int i1;
              print_string ", ";
              print_nonterm t2;
              print_string ")"
         | Min(t1,t2) ->
              print_string "min(";
              print_nonterm t1;
              print_string ", ";
              print_nonterm t2;
              print_string ")"
         | NonKnuthFun(t,_,_) ->
              print_string "non_knuth_fun(";
              print_nonterm t;
              print_string ")"
         | KnuthFun(t,f) ->
              print_string "knuth_fun(";
              print_nonterm t;
              print_string ")"

    let print_eq eq =
       print_nonterm eq.lhs;
       print_string " <- ";
       print_term eq.rhs;
       print_newline()

    let print_eqlist k =
       open_hovbox 0;
       List.iter (fun nonterm -> List.iter print_eq nonterm.l_eqlist) k;
       close_box()

    let solve k = 
       knuth_reset k; (* This is necessary only because of NonKnuthFun
			 that may appear in a scc.
			 We must delete them and clear the associated 
			 dependencies, before computing the solution *)
       let q = KnuthPriorityQueue.create k in
       if !Control.show_equations then
          Message.debug 1 (fun () -> print_eqlist k);
       try
          while true do
             let nonterm = KnuthPriorityQueue.extractmax q in
                nonterm.index <- solved;
                List.iter (fun eq -> 
                    eq.count <- eq.count-1;
                    if eq.count = 0 then KnuthPriorityQueue.update q eq.lhs
                    (compute eq.rhs)) nonterm.r_eqlist;
                nonterm.r_eqlist <- [];
                nonterm.l_eqlist <- []
          done
       with
          KnuthPriorityQueue.EmptyQueue -> ()

    let rec depth_search1 accu nonterm =
       if nonterm.index = white then
       begin
          nonterm.index <- gray;
          nonterm :: (List.fold_left (fun accu eq -> depth_search1 accu eq.lhs)
             accu nonterm.r_eqlist)
       end
       else
          accu

    let rec depth_search2 accu nonterm =
       if nonterm.index = gray then
       begin
          nonterm.index <- black;
          nonterm :: (List.fold_left depth_term accu nonterm.l_eqlist)
       end
       else
          accu

    and depth_term accu eq = match eq.rhs with
           Term _ -> accu
         | Nonterm nonterm -> depth_search2 accu nonterm
         | Minc(_,t2) -> depth_search2 accu t2
         | Min(t1,t2) -> depth_search2 (depth_search2 accu t1) t2
         | NonKnuthFun(t,_,_) -> depth_search2 accu t
         | KnuthFun(t,_) -> depth_search2 accu t

    let cfc ()= 
       let k = !syst in
       reset();
       List.iter (fun nonterm -> nonterm.index <- white) k;
       let k' = List.fold_left depth_search1 [] k in
       List.iter (fun nonterm -> let composante = depth_search2 [] nonterm in
         if composante<>[] then solve composante;
         List.iter (fun nonterm -> nonterm.index <- solved) composante) k'

    let cfc = Profile.f1 "KnuthSolve.cfc" cfc

end

