(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                  Alain Deutsch, INRIA Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Types
open Typedtree
open Format
open Fsolver
open Fenv
open Fenvir

type repr_t =
    RTvar
  | RTtuple of repr_t list
  | RTarrow of repr_t * repr_t
  | RTconstr of Path.t * repr_t list
  | RTobject

module TypeMap = Map.Make (struct
                              type t = repr_t

                              let compare (x :t) (y :t) = compare x y
                           end)

module TranslType =
struct

open NewType

let rec subst inst_subst ty =
  let rec copy ty =
    match ty.e_desc with
      ETvar v as t ->
        (try
          List.assq t inst_subst
        with Not_found ->
          Message.error "TranslType.copy" 
           (fun () -> print_string "Type variable not found"))
    | ETarrow(t1, t2) ->
      	annotate_desc (ETarrow(copy t1, copy t2))
    | ETtuple tl ->
      	annotate_desc (ETtuple(List.map copy tl))
    | ETconstr(p, tl) ->
        trdecl p (List.map copy tl)
  in copy ty

and trdecl decl params =
  match decl.e_type_kind with
    EType_manifest body -> subst (List.combine decl.e_type_params params) body
  | _ -> annotate_desc (ETconstr(decl, params))

let declsubstitute params args body =
   let copy = subst (List.combine params (List.map annotate_desc args)) in
   match body with
      EType_abstract -> body
    | EType_group fieldlist -> EType_group (List.map copy fieldlist)
    | EType_manifest ty -> EType_manifest (copy ty)

(* Post treatment of type declarations *)

(* 1. Put -1 as top when in the top scc, -2 otherwise *)

let rec put_top ty =
  match ty.e_desc with
    ETvar _ -> ty.e_top <- -2
  | ETarrow(t1,t2) -> put_top t1; put_top t2; ty.e_top <- t2.e_top
  | ETtuple(types) -> 
      List.iter put_top types;
      if List.exists (fun ty -> ty.e_top = -1) types then
	ty.e_top <- -1
      else
	ty.e_top <- -2
  | ETconstr(decl,types) ->
      List.iter put_top types;
      if decl.e_color <> 4 then ty.e_top <- -1 else
      if List.exists (fun ty -> ty.e_top = -1) types then
	ty.e_top <- -1
      else
	ty.e_top <- -2

let decl_nlp decl = 
   let rec type_nlp ty =
      match ty.e_desc with
           ETvar(_) -> false
         | ETtuple(types) -> List.exists type_nlp types
         | ETarrow(_,_) -> true
         | ETconstr(decl,types) -> if (decl.e_color <> 4) then false else
              decl.e_bad_type or decl.e_nlp_decl or (match decl.e_type_kind with
                EType_abstract -> false (* Optimisation - 19/11/96 *)
              | _ -> List.exists type_nlp types)
   in match decl.e_type_kind with
         EType_abstract -> false (* Optimisation - 19/11/96 *)
       | EType_group fieldlist -> List.exists type_nlp fieldlist
       | EType_manifest ty -> type_nlp ty

(* 2. Calcul des composantes fortement connexes *)

(* 2.1. Parcours en profondeur dans le sens direct *)

let rec depth1 accu info =
  if info.e_color = 0 then
  begin
     info.e_color <- 1;
     info :: (match info.e_type_kind with
          EType_abstract -> accu
        | EType_manifest e -> makedep info accu e
        | EType_group(fieldlist) -> List.fold_left 
             (makedep info) accu fieldlist)
  end
  else
    accu

and makedep info accu ty = match ty.e_desc with
    ETvar _ -> accu
  | ETarrow(_,t2) -> makedep info accu t2
  | ETtuple(types) -> List.fold_left (makedep info) accu types
  | ETconstr(decl, types) ->
       if decl.e_color <> 4 then
           decl.e_inverse_dep <- info :: decl.e_inverse_dep;
       List.fold_left (makedep info) (depth1 accu decl) types

(* 2.2. Parcours en profondeur dans le sens inverse *)

let rec depth2 accu info =
  if info.e_color = 1 then
  begin
     info.e_color <- 2;
     info :: (List.fold_left depth2 accu info.e_inverse_dep)
  end
  else
     accu

(* 2.3. Calcul des composantes fortement connexes = deux parcours en profondeur
   cf Cormen, Leiserson, Rivest *)

let cfc decllist =
   let decllist' = List.fold_left depth1 [] decllist in
      List.map (depth2 []) decllist'

(* 3. Verification de la correction des dependances *)

exception Baddep

let rec makedep ty = match ty.e_desc with
    ETvar _ -> ()
  | ETarrow(_,t2) -> makedep t2
  | ETtuple(types) -> List.iter makedep types
  | ETconstr(decl, types) -> 
       if decl.e_color = 2 then
       begin
          let args = List.map (fun arg -> arg.e_desc) types in
             List.iter (function ETvar _ -> ()
                               | _ -> raise Baddep) args;
             decl.e_type_kind <- declsubstitute decl.e_type_params args
                decl.e_type_kind;
             decl.e_type_params <- args;
             decl_makedep decl
       end
       else if decl.e_color = 3 then
       begin
          List.iter2 (fun ty arg -> if ty.e_desc <> arg then raise Baddep)
            types decl.e_type_params
       end
       else
          List.iter makedep types

and decl_makedep decl = 
     decl.e_color <- 3; 
     match decl.e_type_kind with
          EType_abstract -> ()
        | EType_manifest e -> makedep e
        | EType_group(fieldlist) -> List.iter makedep fieldlist

(* 5. Function to treat one strongly connected component *)

let decl2 decllist =
  let (dec1 :: decll) = decllist in

(* 5.1 Check for bad dependances and change the declarations so that the 
variables are the same in all the declarations, not changed by substitutions *)

  decl_makedep dec1;
  (* Just for safety *)
  List.iter (fun decl -> if decl.e_color <> 3 then
     Message.error "decl2" (fun () -> 
        print_string "This is not a strongly connected component")) decllist;

(* 5.3 Checks if the declarations are level-preserving *)

   let nlp = List.exists decl_nlp decllist in
   List.iter (fun decl -> decl.e_nlp_decl <- nlp) decllist;   

(* 5.4 Compute the graph *)

   (* Put -1 as top when in the top scc, -2 otherwise *)
   List.iter (fun decl -> match decl.e_type_kind with
     EType_abstract -> ()
   | EType_manifest ty -> put_top ty
   | EType_group tyl -> List.iter put_top tyl) decllist;
   List.iter (fun decl -> decl.e_type_graph <- decllist) decllist

(* 6. Post treatment of type declarations : main program.
      Splits into strongly connected components and then treat each
      component one after the other. *)

  let rec expand ty =
    match ty.e_desc with
      ETvar _ -> ty
    | ETarrow(ty1,ty2) -> annotate_desc(ETarrow(expand ty1, expand ty2))
    | ETtuple(types) -> annotate_desc (ETtuple(List.map expand types))
    | ETconstr(decl, args) ->
	trdecl decl (List.map expand args)

  let expanddecl decl =
    match decl.e_type_kind with
      EType_abstract -> ()
    | EType_manifest ty -> decl.e_type_kind <- EType_manifest (expand ty)
    | EType_group tyl -> decl.e_type_kind <- EType_group (List.map expand tyl)

let posttreatment decllist =
   List.iter expanddecl decllist;
   let composantes = cfc decllist in
   List.iter (fun decllist -> 
      if decllist <> [] then
      begin
         begin
            try
               decl2 decllist
            with Baddep ->
               List.iter (fun decl ->
                  decl.e_bad_type <- true;
                  Message.debug 0 (fun () ->
                     print_string "The type ";
                     Printtyp.ident decl.e_name;
                     print_string " has bad dependances.")) decllist
         end;
         List.iter (fun decl -> decl.e_color <- 4) decllist
      end
      ) (List.rev composantes);
  List.iter (fun decl -> decl.e_inverse_dep <- []) decllist

(* Global program to translate type declarations *)

let decl_object = {e_modname = "";
                   e_name = Ident.create "object";
                   e_type_params = [];
                   e_color = 4;
                   e_inverse_dep = [];
                   e_type_kind = EType_abstract;
                   e_bad_type = false;
		   e_nlp_decl = false;
                   e_type_graph = []}

let desc_object = ETconstr(decl_object, [])

let rec notdistinct = function [] -> false
                             | (a::l) -> (List.memq a l) || notdistinct l

let modname = ref ""

let counter = ref 0

let decl env decllist =
   let instsubst = ref [] in
   (* Create the (empty) declarations *)
   let (env, decllist') = List.fold_right (fun (name, info) (env, decllist') ->
      let result = {e_modname = !modname;
                    e_name = name;
                    e_type_params = List.map (fun v -> let v = Ctype.repr v in
                       try 
                          List.assq v !instsubst (* Already used variable *)
                       with Not_found ->
                          let x = ETvar (!counter) 
                          in instsubst := (v,x) :: !instsubst;
                             incr counter; x) info.type_params;
                    e_type_kind = EType_abstract;
                    e_color = 0;
                    e_inverse_dep = [];
                    e_bad_type = false;
                    e_nlp_decl = false;
                    e_type_graph = []}
      in (Envir.add_type name result env, result :: decllist')) decllist (env, []) in
   (* Translate a type expression of a declaration *)
   let rec expr ty =
        { e_desc = 
           (let ty = Ctype.repr ty in
            match ty.desc with
               Tvar -> (try List.assq ty !instsubst with Not_found ->
                  Message.error "TranslType.decl.expr" 
                  (fun () -> print_string "Type variable not found"))
             | Tarrow(ty1,ty2) -> ETarrow(expr ty1, expr ty2)
             | Ttuple(types) -> ETtuple(List.map expr types)
             | Tconstr(name, args, _) ->
                  ETconstr(Envir.lookup_type name env, List.map expr args)
             | Tnil | Tfield _ ->
                  Message.error "Transltype.decl" (fun () -> print_string
                   "Object field cannot be in place of a type")
             | Tobject _ -> desc_object (* TO DO *)
             | Tlink _ ->
                  Message.error "Transltype.decl" (fun () -> print_string
                   "Tlink should have been removed by Ctype.repr"));
	  e_id = 0;
          e_top = -1;
          e_nlp = true} in
   (* Translate the declarations and put them in the empty declarations
      created above *)
   List.iter2 (fun (_,info) decl' -> 
     let type_vars = List.map Ctype.repr info.type_params in
     if (List.exists (fun v -> v.desc <> Tvar) type_vars) 
        || notdistinct type_vars
     then
     begin
        decl'.e_bad_type <- true;
        Message.debug 0 (fun () ->
                     print_string "The type ";
                     Printtyp.ident decl'.e_name;
                     print_string " has bad dependances.")
     end
     else
        decl'.e_type_kind <-
     match info.type_manifest with
       Some e -> EType_manifest (expr e)
     | None -> match info.type_kind with
                 Type_abstract -> EType_abstract
               | Type_variant(varlist) -> EType_group (List.fold_left 
                    (fun accu (_,explist) -> List.fold_left (fun accu exp
                    -> (expr exp) :: accu) accu explist) [] varlist)
(* This is ok since we cannot extract the tuples of each variant.
   In objective Caml, we can extract them, so we must consider the
   variant as a group of tuples !! 
      EType_group (List.map (fun (_,explist) -> ETtuple(List.map expr
      explist) varlist)                                                *)
               | Type_record(fieldlist) -> EType_group  (List.map 
                    (fun (_,_,exp) -> expr exp) fieldlist)) decllist decllist';
   (* Compute graph, top and nlp for each declaration *)
   posttreatment decllist';
   env

(* Instantiation of type declarations when applying a functor *)

open TEnv

(* List of type declarations for e_inverse_dep field has been set
   to remember associations between types before and after functor
   application *)
let cleanup_list = ref []

let rec matchtypes randsig randmod = 
    match (randsig, randmod) with
        (Module mrandsig, Module mrandmod) ->
           Ident.iter (fun typename typesig -> 
	     cleanup_list := typesig :: (!cleanup_list);
             typesig.e_inverse_dep <- [Ident.find_name (Ident.name
                 typename) mrandmod.types]) mrandsig.types;
           Ident.iter (fun modname modsig ->
             matchtypes modsig (Ident.find_name (Ident.name modname)
                 mrandmod.submod)) mrandsig.submod;
           Ident.iter (fun modname modsig ->
             matchtypes modsig (Ident.find_name (Ident.name modname)
                 mrandmod.submod_types)) mrandsig.submod_types
      | (Functor(sig1, sig2), Functor(mod1, mod2)) ->
           matchtypes sig1 mod1;
           matchtypes sig2 mod2
      | (_, _) -> Message.error "matchtypes" (fun () -> print_string
           "Argument of functor does not correspond to signature")

let ident_fold fn accuanc accunouv tbl =
   let x = ref [] in
   Ident.iter (fun name id -> x := (name, id) :: (!x)) tbl;
   List.fold_left (fun (accu, accuanc, accunouv) (name, id) -> 
     let (id', accuanc', accunouv') = fn accuanc accunouv id in
     (Ident.add name id' accu, accuanc', accunouv'))
     (Ident.empty, accuanc, accunouv) (!x)

let copy_types1 accuanc accunouv typetbl =
   let typelist = ref [] in
   Ident.iter (fun name id -> typelist := id :: (!typelist)) typetbl;
   let (env, accuanc', accunouv') = List.fold_left (fun (env, tylist, decllist') info ->
      match info.e_inverse_dep with
        [result] -> 
          (Ident.add result.e_name result env, tylist, decllist')
      | [] ->
	  let result = 
	    { e_modname = info.e_modname;
	      e_name = Ident.create (Ident.name info.e_name);
              e_type_params = info.e_type_params;
              e_type_kind = EType_abstract;
              e_color = 0;
              e_inverse_dep = [];
              e_bad_type = false;
              e_nlp_decl = false;
              e_type_graph = []}
	  in
	  cleanup_list := info :: (!cleanup_list);
	  info.e_inverse_dep <- [result];
	  (Ident.add result.e_name result env, info :: tylist, result :: decllist')
      | _ -> Message.error "copy_types" (fun () -> print_string
          "inverse_dep not cleaned up"))
     (Ident.empty, accuanc, accunouv) (!typelist) in
   (env, accuanc', accunouv')

let translate accuanc accunouv =
   let rec expr ty =
        { e_desc =
           begin
           match ty.e_desc with
               ETvar _ -> ty.e_desc
             | ETarrow(ty1,ty2) -> ETarrow(expr ty1, expr ty2)
             | ETtuple(types) -> ETtuple(List.map expr types)
             | ETconstr(decl, args) ->
                  ETconstr((match decl.e_inverse_dep with
                    [] -> decl
                  | [decl'] -> decl'
                  | _ -> Message.error "translate" (fun () -> print_string
                     "inverse_dep not cleaned up")), List.map expr args)
           end;
	  e_id = 0;
          e_top = -1;
          e_nlp = true} in
   List.iter2 (fun ancdecl nouvdecl ->
      nouvdecl.e_type_kind <- match ancdecl.e_type_kind with
        EType_abstract -> EType_abstract
      | EType_group typelist -> EType_group (List.map expr typelist)
      | EType_manifest exp -> EType_manifest (expr exp)) accuanc accunouv;
   List.iter (fun decl -> decl.e_inverse_dep <- []) (!cleanup_list);
   cleanup_list := [];
   posttreatment accunouv

let rec copy_types accuanc accunouv = function
    Module m -> 
       let (types', accuanc', accunouv') = 
          copy_types1 accuanc accunouv m.types in
       let (submod', accuanc'', accunouv'') = 
          ident_fold copy_types accuanc' accunouv' m.submod in
       let (submod_types', accuanc''', accunouv''') =
          ident_fold copy_types accuanc'' accunouv'' m.submod_types in
       (Module {env = m.env; types = types'; submod = submod';
                submod_types = submod_types'}, accuanc''', accunouv''')
  | Functor(sig1, sig2) ->
      (* If we have a functor functor(R1 : S1) -> functor(R2 : S2) -> ...
	 S2 may refer to types in S1. Therefore, when applying the
	 functor to its first parameter M1, the types in S1 must be 
	 substituted in S2 by types in M1. Hence copy_types
	 must also by applied to the parameter S2 of the functor 
	 (here S2 <-> sig1) *)
       let (sig1', accuanc', accunouv') = copy_types accuanc accunouv sig1 in
       let (sig2', accuanc'', accunouv'') = copy_types accuanc' accunouv' sig2
       in (Functor(sig1', sig2'), accuanc'', accunouv'')

let instantiate rand_var randenv body =
    matchtypes rand_var randenv;
    let (body', accuanc, accunouv) = copy_types [] [] body in
    translate accuanc accunouv;
    body'

let copy_sig body =
    let (body', accuanc, accunouv) = copy_types [] [] body in
    translate accuanc accunouv;
    body'


(* Translate type expressions to the simpler language *)
type typemap_t = NewType.t_type_expr TypeMap.t

let typemap = ref (TypeMap.empty : typemap_t)
  (* This does not memorize submodules definitions : when we close a
     module, we forget all the definitions in this module because the
     names of the declarations change when we change the current module *)

(* Using this function instead of Ctype.repr avoids the segmentation
   fault when compiling typing/mtype.ml. Why ? Perhaps a recursive 
   abbreviation in the abbrev_memo of a TConstr, which would cause
   an infinite loop in the comparison ?

   It should speed up the work of TranslType.expr, by enabling TypeMap.find
   to succeed more often (types which differed only by the presence of Tlink
   or by their level are now identical). 
*)


let rec repr ty = 
   match ty.desc with
     Tlink t -> repr t
   | Tvar -> RTvar
   | Ttuple l -> RTtuple (List.map repr l)
   | Tarrow(p,d) -> RTarrow (repr p, repr d)
   | Tconstr(p, params, _) -> RTconstr(p, List.map repr params)
   | Tobject _ -> RTobject (* Do not treat object types *)
   | Tfield _ | Tnil -> Message.error "TranslType.repr" (fun () -> ())
       (* There should not be object fields outside an object type *)


let trdecl decl params =
  match decl.e_type_kind with
    EType_manifest body -> Subst.substitute2 decl.e_type_params params body
  | _ -> Subst.trdecl2 decl params

let et_v = Subst.annotate_htbl (ETvar 0) 1 false
let et_obj = Subst.annotate_htbl desc_object 1 false

let expr env ty =
  let rec expr' ty =
    try
      TypeMap.find ty !typemap
    with Not_found ->
      let result =
        match ty with
          RTvar -> et_v
        | RTarrow(ty1,ty2) -> 
	    let dst = expr' ty2 in
	    Subst.annotate_htbl (ETarrow(expr' ty1, dst)) dst.e_top true
        | RTtuple(types) -> 
	    let tyl = List.map expr' types in
	    Subst.annotate_htbl (ETtuple tyl) (1 + max_list tyl) 
	      (List.exists (fun ty -> ty.e_nlp) tyl)
        | RTconstr(name, args) -> 
            trdecl (Envir.lookup_type name env) (List.map expr' args)
        | RTobject -> et_obj (* TO DO *)
      in 
      typemap := TypeMap.add ty result !typemap;
      result
  in 
  expr' (repr ty)

let expr = Profile.f2 "TranslType.expr" expr

end

