(***********************************************************************)
(*                                                                     *)
(*              Fast Escape Analysis - Environment                     *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open L1solver

val list_map3 :
  ('a -> 'b -> 'c -> 'd) -> 'a list -> 'b list -> 'c list -> 'd list
val list_fold_max : ('a -> 'b -> int) -> 'a list -> 'b list -> int

module NewType :
  sig
    type t_type_declaration =
      { e_modname: string;
        e_name: Ident.t;
        mutable e_type_params: t_type_expr_desc list;
        mutable e_color: int;
        mutable e_inverse_dep: t_type_declaration list;
        mutable e_type_kind: t_type_kind;
        mutable e_bad_type: bool;
    	mutable e_nlp_decl: bool;
        mutable e_type_graph: t_type_declaration list }
    and t_type_kind =
      | EType_abstract
      | EType_group of t_type_expr list
      | EType_manifest of t_type_expr
    and t_type_expr_desc =
      | ETvar of int
      | ETarrow of t_type_expr * t_type_expr
      | ETtuple of t_type_expr list
      | ETconstr of t_type_declaration * t_type_expr list
    and t_type_expr =
      { mutable e_top: int;
        mutable e_nlp: bool;
	e_id: int;
        e_desc: t_type_expr_desc }
    val max_list : t_type_expr list -> int
    val print_with_delimitors : ('a -> unit) -> string -> 'a list -> unit
    val print : t_type_expr -> unit
    val same_decl : t_type_declaration -> t_type_declaration -> bool
    val equal : t_type_expr -> t_type_expr -> bool
    val functional : t_type_expr -> bool
    val extract_types_of_parameters :
      bool -> t_type_expr -> int -> t_type_expr list * t_type_expr
    val annotate_desc : t_type_expr_desc -> t_type_expr
    val hash : t_type_expr -> int
  end

module Subst :
  sig
    open NewType
    val annotate_htbl : t_type_expr_desc -> int -> bool -> t_type_expr
    val trdecl2 : t_type_declaration -> t_type_expr list -> t_type_expr
    val substitute2 : 
      t_type_expr_desc list -> t_type_expr list -> t_type_expr -> t_type_expr
  end

module Lattice :
  sig
    type t = int
    val top : int
    val topof : NewType.t_type_expr -> t
    val not_level_preserving : NewType.t_type_expr -> bool
    val typeinfo : NewType.t_type_expr -> t
    val print_as_type_annotation : t -> NewType.t_type_expr -> unit
    val is_ident : NewType.t_type_expr -> NewType.t_type_expr -> bool
    val inst_gen : NewType.t_type_expr -> NewType.t_type_expr -> t -> t
    val min_inst : NewType.t_type_expr -> NewType.t_type_expr -> int -> int
  end

module Context :
  sig
    type op =
      { seuil: KnuthEquations.term;
        supimp: KnuthEquations.term;
        infimp: KnuthEquations.term;
        func: KnuthEquations.term;
        infimpr: KnuthEquations.term }
    type elemnode
    type t = (elemnode * NewType.t_type_expr * NewType.t_type_expr) list ref
    val opid : op
    val opzero : op
    val optail : Lattice.t -> op
    val optop : Lattice.t -> op
    val optopinter : bool -> Lattice.t -> op
    val opseuil : Lattice.t -> Lattice.t -> op
    val newnode : unit -> elemnode
    val newop : string -> op
    val genterm : op -> NewType.t_type_expr -> NewType.t_type_expr -> op
    val addeq : op -> op -> unit
    val eval : elemnode -> op
    val link : elemnode -> elemnode -> op -> unit
    val register_anal : op -> t -> unit
    val register_anal0 : t -> op
  end

module CTree :
  sig
    type op = { c_seuil: int; c_func: int; c_supimp: int; c_infimp: int }
    type letone =
      Asttypes.rec_flag * Asttypes.stack_flag * Lattice.t *
      (Typedtree.pattern * Typedtree.expression) list
    type blocks = (op * lettree) list
    and lettree = | LetGroup of blocks | LetOne of letone
    val opzero : op
    val opid : op
    val getvalue : op -> Lattice.t -> Lattice.t -> bool
    val eval : blocks -> op * letone * blocks
    val clearone : blocks -> blocks
    val compose : op -> op -> op
  end

module Store :
  sig
    type op =
      { s_seuil: KnuthEquations.term;
        s_func: KnuthEquations.term;
        s_supimp: KnuthEquations.term;
        s_infimp: KnuthEquations.term }
    val newop : string -> op
    val to_norm : bool -> op -> Lattice.t -> Context.op
    val opid : op
    val opzero : op
    val opcst : int -> op
    val opinter : int -> op
    val opseuil : int -> int -> op
    val opseuilinter : int -> int -> op
    val opnlp : op
    val get_norm : Context.op -> op
    val get_tail : bool -> Context.op -> op
    val to_const : op -> CTree.op
    val getvalue : op -> int -> int
    val instterm :
      (KnuthEquations.term -> KnuthEquations.term) ->
      op -> NewType.t_type_expr -> NewType.t_type_expr -> op
    val addeq : op -> Context.op -> unit
    val register_anal : op -> Context.t -> unit
    val compose : op -> op -> op
  end

module VarEnv :
  sig
    type t = Context.t Ident.tbl
    val create : unit -> t
    val update : Ident.t -> Context.t -> t -> Context.t Ident.tbl
    val lookup : Ident.t -> t -> Context.t option
  end

type expression =
  { e_desc: expres_desc;
    e_loc: Location.t;
    e_type: Types.type_expr;
    e_env: Env.t;
    e_tail: bool;
    e_tailcalls: bool;
    e_unknown_tailcalls: bool;
    e_node: Context.elemnode;
    e_tye: NewType.t_type_expr;
    e_type_info: Lattice.t;
    mutable e_immop: Store.op;
    mutable e_stackop: Store.op;
    mutable e_stackty: Lattice.t;
    mutable e_stacktailop: Store.op;
    mutable e_stacktailty: Lattice.t }

and expres_desc =
  | Eexp_ident of Path.t * Types.value_description
  | Eexp_constant of Asttypes.constant
  | Eexp_let of Asttypes.rec_flag * Asttypes.stack_flag *
                  (Typedtree.pattern * expression) list * expression
  | Eexp_function of (Typedtree.pattern * expression) list
  | Eexp_apply of expression * expression list
  | Eexp_tuple_apply of expression * expression list
  | Eexp_match of expression * (Typedtree.pattern * expression) list
  | Eexp_try of expression * (Typedtree.pattern * expression) list
  | Eexp_tuple of expression list
  | Eexp_construct of Types.constructor_description * expression list
  | Eexp_record of (Types.label_description * expression) list
  | Eexp_field of expression * Types.label_description
  | Eexp_setfield of expression * Types.label_description * expression
  | Eexp_array of expression list
  | Eexp_ifthenelse of expression * expression * expression option
  | Eexp_sequence of expression * expression
  | Eexp_while of expression * expression
  | Eexp_for of Ident.t * expression * expression * Asttypes.direction_flag *
                  expression
  | Eexp_when of expression * expression
  | Eexp_send of expression * string
  | Eexp_new of Path.t
  | Eexp_instvar of Path.t * Path.t
  | Eexp_setinstvar of Path.t * Path.t * expression
  | Eexp_override of Path.t * (Path.t * expression) list

module IdentInfo :
  sig
    type t =
      { type_expr: NewType.t_type_expr;
        tupled: bool;
        mutable current_scc: bool;
        mutable unknown_tail_call: bool;
        escape_info: Store.op list }
    val escape_info : t -> Store.op list
    val type_expr : t -> NewType.t_type_expr
    val arity : t -> int
    val empty : NewType.t_type_expr -> bool -> t
    val according_exp : Typedtree.expression -> bool * Store.op list
    val according_Eexp : expression -> bool * Store.op list
    val build : NewType.t_type_expr -> bool * Store.op list -> bool -> t
    val print : t -> unit
    val polymorphic_instantiation :
      bool -> t -> NewType.t_type_expr -> Context.op list
    val lookup_primitive : string -> NewType.t_type_expr -> t
    val box_float : string -> bool
    val is_an_allocation : string -> bool
  end

module TEnv :
  sig
    type t_env = IdentInfo.t Ident.tbl
    type t_typedecl = NewType.t_type_declaration Ident.tbl
    type t_hierarchy =
      { env: t_env;
        types: t_typedecl;
        submod: t_submodules;
        submod_types: t_submodules }
    and t_hierarchy_mod =
      | Module of t_hierarchy
      | Functor of t_hierarchy_mod * t_hierarchy_mod
    and t_submodules = t_hierarchy_mod Ident.tbl
    val emptyhierarchy : t_hierarchy
  end

