(***********************************************************************)
(*                                                                     *)
(*              Fast Escape Analysis - Equations solver                *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                  Alain Deutsch, INRIA Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Typedtree
open Format

let top = 32000

module KnuthEquations =
struct

type term = Term of int
          | Nonterm of nonterm
          | Min of term * term
          | Max of term * term
          | Seuil of term * term * term
            (* Seuil(t1,t2,t3) = if t1 >= - t2 then t3 else -top
               le - devant t2 permet que plus t2 est petit plus l'info est
               precise, donc on cherche un plus petit point fixe pour t2
               aussi *)
          | Fun of term * (int -> int)

and equation = {lhs : nonterm; mutable rhs : term}

and data = equation list

and t = {mutable key : int;
         mutable index : int;
         mutable l_eqlist : data;
         mutable r_eqlist : data;
         name : string}

and nonterm = t;;

(* Current value * equations having this non terminal as lhs * 
   equations using this non terminal on their rhs * index * name

   index is 0 when building the equations, >=0 when solving them,
   and -1 when equations are solved, ie when the current value won't change
   any more *)

end

module KnuthSolve :
sig 

open KnuthEquations
val reset : unit -> unit
    (* Resets the system *)
val newnonterm : string -> nonterm
    (* [newnonterm name] creates a new non terminal *)
val compute : term -> int
val addeq : nonterm -> term -> unit
    (* [addeq lhs rhs] adds an equation *)
val reduce : term -> term
val termmin : term -> term -> term
val termmax : term -> term -> term
val nonknuthfun : term -> (int -> int) -> int -> term
    (* [nonknuthfun t f kf] creates a term whose value is f(t) except
       if it is in a cycle, where we cannot use non Knuth functions,
       and where we take the constant approximation kf. We must have
       kf >= max { f(t) | t term } for correctness *)
val knuthfun : term -> (int -> int) -> term
    (* [knuthfun t f] creates a term whose value is f(t).
       We must have f(t) <= t to be able to apply the Knuth algorithm *)
val print_term : term -> unit
val cfc : unit -> unit
    (* Solves the system *)
end =
struct

open KnuthEquations

(* two phases :

1. maintain a list of non-terminals with a list of the equations which use it 
   on their right hand-side, and the current key

2. solve the system by the Knuth algorihtm, after building a priority queue
   from the list mentionned above
*)

let syst = ref []

let reset() = (syst := [])

let solved = -1
let black = -2
let gray = -3
let white = -4

let newnonterm name = let nonterm = {key = -top; l_eqlist = []; r_eqlist = []; index = 0; name = name} in
          syst:= nonterm:: !syst;
          nonterm

let rec simplifterm = function
     (Term _) as t -> t
   | (Nonterm nonterm) as t -> if nonterm.index = solved then
                    Term nonterm.key
                 else
                    t
   | Max(t1,t2) -> 
                 begin
                   match (simplifterm t1, simplifterm t2) with
                     (Term i1, Term i2) -> Term(max i1 i2)
                   | (Term i1, x2) when i1 = -top -> x2 
                   | (x1, Term i2) when i2 = -top -> x1
                   | (x1,x2) -> Max (x1,x2)
                 end
   | Min(t1,t2) -> 
                 begin
                   match (simplifterm t1,simplifterm t2) with
                     (Term i1, Term i2) -> Term(min i1 i2)
                   | (Term i1,x2) -> if i1 = top then x2 else 
                                     if i1 = -top then Term (-top) else
                                        Min(Term i1, x2)
                   | (x1,Term i2) -> if i2 = top then x1 else
                                     if i2 = -top then Term (-top) else 
                                        Min(Term i2, x1)
                   | (t1,t2) -> Min(t1,t2)
                 end
   | Seuil(t,t',t1) ->
                 begin
                   match simplifterm t' with
                     Term 0 -> simplifterm t1
                   | Term n -> if n = -top then Term (-top)
                               else begin
                        match simplifterm t with
                           Term i -> if i >= -n then simplifterm t1 else Term (-top)
                         | t -> Seuil(Term n, t,simplifterm t1)
                        end
                   | t' -> Seuil(simplifterm t, t', simplifterm t1)
                 end
   | Fun(t1,f) -> 
                 match simplifterm t1 with 
                     Term i1 -> Term (f i1)
                   | t1 -> Fun(t1,f)

let addeq_reqlist eq t =
     let rec phi = function
              Term _ -> ()
            | Nonterm nonterm -> if nonterm.index <> solved then
                    nonterm.r_eqlist <- eq :: nonterm.r_eqlist
            | Max(t1,t2) -> phi t1; phi t2
            | Min(t1,t2) -> phi t1; phi t2
            | Seuil(t,t',t1) -> phi t; phi t'; phi t1
            | Fun(t1,f) -> phi t1
     in phi t

let rec compute = function
              Term x -> x
            | Nonterm nonterm -> nonterm.key
            | Max(t1,t2) -> max (compute t1) (compute t2)
            | Min(t1,t2) -> min (compute t1) (compute t2)
            | Seuil(t,t',t1) -> if (compute t)>= - (compute t') then 
                    compute t1 
                 else -top
            | Fun(t1,f) -> f (compute t1)

let print_nonterm {name = name; index = index} =
       print_string name;
       print_int index

let rec print_term = function
           Term n -> print_int n
         | Nonterm nonterm -> 
              if nonterm.index = solved then print_int nonterm.key else
                 print_nonterm nonterm
         | Max(t1,t2) ->
              print_string "max(";
              print_term t1;
              print_string ", ";
              print_term t2;
              print_string ")"
         | Min(t1,t2) ->
              print_string "min(";
              print_term t1;
              print_string ", ";
              print_term t2;
              print_string ")"
         | Seuil(t, n, t1) ->
              print_string "(if ";
              print_term t;
              print_string " >= - ";
              print_term n;
              print_string " then ";
              print_term t1;
              print_string " else -top)";
         | Fun(t,_) ->
              print_string "fun(";
              print_term t;
              print_string ")"

let print_eq eq =
       print_nonterm eq.lhs;
       print_string " <- ";
       print_term eq.rhs;
       print_newline()

let addeq_nosimplif lhs rhs = 
       let eq = {lhs = lhs; rhs = rhs} in
       addeq_reqlist eq rhs;
       match eq.rhs with
               Term res1 -> lhs.key <- max lhs.key res1 
             | _ -> lhs.l_eqlist <- eq :: lhs.l_eqlist


let addeq lhs rhs = addeq_nosimplif lhs (simplifterm rhs)

let reduce t = match simplifterm t with 
                      (Nonterm _ | Term _) as t -> t
                    | t -> let t' = newnonterm "tont" in
                           addeq_nosimplif t' t;
                           Nonterm t'

let termmin t1 t2 = Min(t1, t2)

let termmax t1 t2 = Max(t1, t2)

let nonknuthfun t f kf = Fun(t, f)

let knuthfun t f = Fun(t, f)


let print_eqlist k =
       open_hovbox 0;
       List.iter (fun nonterm -> List.iter print_eq nonterm.l_eqlist) k;
       close_box()

let rec oneiter k =
       let changed = ref false in
       List.iter (fun nonterm -> List.iter 
          (fun eq -> let rhs = compute eq.rhs in
                     if rhs > eq.lhs.key then
                     begin
                        changed := true;
                        eq.lhs.key <- rhs
                     end) nonterm.l_eqlist) k;
       if !changed then oneiter k

let solve k = 
       if !Control.show_equations then
          Message.debug 1 (fun () -> print_eqlist k);
       oneiter k

let rec depth_search1 accu nonterm =
       if nonterm.index = white then
       begin
          nonterm.index <- gray;
          nonterm :: (List.fold_left (fun accu eq -> depth_search1 accu eq.lhs)
             accu nonterm.r_eqlist)
       end
       else
          accu

let rec depth_search2 accu nonterm =
       if nonterm.index = gray then
       begin
          nonterm.index <- black;
          nonterm :: (List.fold_left depth_term accu nonterm.l_eqlist)
       end
       else
          accu

and depth_term accu eq = depth_term2 accu eq.rhs

and depth_term2 accu = function
           Term _ -> accu
         | Nonterm nonterm -> depth_search2 accu nonterm
         | Max(t1,t2) -> depth_term2 (depth_term2 accu t1) t2
         | Min(t1,t2) -> depth_term2 (depth_term2 accu t1) t2
         | Seuil(t,t',t1) -> depth_term2 (depth_term2 (depth_term2 accu t) t') t1
         | Fun(t,_) -> depth_term2 accu t

let cfc ()= 
       let k = !syst in
       reset();
       List.iter (fun nonterm -> nonterm.index <- white) k;
       let k' = List.fold_left depth_search1 [] k in
       List.iter (fun nonterm -> let composante = depth_search2 [] nonterm in
         if composante<>[] then solve composante;
         List.iter (fun nonterm -> nonterm.index <- solved;
                                   nonterm.l_eqlist <- [];
                                   nonterm.r_eqlist <- []) composante) k'

let cfc = Profile.f1 "KnuthSolve.cfc" cfc

end
