(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)


open Format
open Ttype
open Tsolver
open Tprinttyp

let get = Graph.get
let dummynode = Graph.newnode (Expanded [| |]) None

(**** Call ****)


let sons_corresp = ref []

type status_t = Res | Param
let status = ref Res

let typeinst_accu = ref []

let get_stamp ty = match ty.b_desc with
  BTvar -> (0, tuple_name)
| BTbadType -> (1, tuple_name)
| BTabstract (d, _, _) -> (2, d)
| BTarrow _ -> (3, tuple_name)
| BTgroup(d,l) -> (4 + List.length l, d)
      (* The name of the declaration is not enough to determine 
	 the number of parameters because we have a single
	 declaration for tuples *)

let ty_compare ty1 ty2 = compare (get_stamp ty1) (get_stamp ty2)

module Ord =
struct
  type t = Ttype.graph_t * status_t * Ttype.newtype_t

  let compare (_,_,(t1,_)) (_,_,(t2,_)) = 
    let r = ty_compare t1 t2 in 
    if r = 0 then 1 else r
end

module Orderedset = Ordset.Make(Ord)

let rec instcall tylist = 
  if tylist <> [] then
  if List.for_all (fun (_,_,(ty,_)) -> ty.b_mark = 0) tylist then () else
  begin
    List.iter (fun (_,_,(ty,_)) -> ty.b_mark <- 0) tylist;
  (* tylist = [ gen_node * status * inst_type ] *)
  let all_list = ref [] in
  let s = ref Orderedset.empty in
  (* Types are ordered according to their stamp :
     0 -> BTvar, 1 -> BTbadType, 2 -> BTabstract, 3 -> BTarrow, 4+nbre param -> BTgroup
     * t_type_name *)
  List.iter (fun ((_,_,(ty,_)) as info) ->
    match ty.b_desc with
      BTAll -> all_list := info :: !all_list
    | _ -> s := Orderedset.add info !s) tylist;
  (* call_accu = [ inst_node * gen_node param ] * [ inst_node * gen_node res ]
     * [ [ ty_sons ] ] *)
  let rec add_all nsons = function
      [] -> ([], [], Misc.replicate_list [] nsons)
    | (((gen_node, status, (_, inst_nt)) as info)::l) ->
	let (paramcorresp, rescorresp, sons) = add_all nsons l in
	let newsons = List.map (fun l -> info :: l) sons in
	if status = Res 
	then (paramcorresp, (inst_nt, gen_node)::rescorresp, newsons)
	else ((inst_nt, gen_node)::paramcorresp, rescorresp, newsons)
  in
  let call_accu = add_all 
      (let (_, _, (firstty,_)) =  Orderedset.first !s in
      match firstty.b_desc  with
	  BTvar | BTabstract _ | BTbadType -> 0
	| BTarrow _ -> 1
	| BTgroup(_, l) -> List.length l) !all_list
  in
  let ((paramcorresp, rescorresp, sons), _) = Orderedset.fold 
      (fun (gen_node, status, (inst_type, inst_nt)) 
	   ((paramcorresp, rescorresp, sons) as call_accu, last_type)
	    ->
    let (paramcorresp, rescorresp, sons) = 
    if match last_type with
      None -> true
    | Some last_ty -> ty_compare last_ty inst_type == 0 then call_accu
    else
      begin
      	Graph.call paramcorresp [] rescorresp [];
	List.iter instcall sons;
	add_all (match inst_type.b_desc with
	  BTvar | BTabstract _ | BTbadType -> 0
	| BTarrow _ -> 1
	| BTgroup(_, l) -> List.length l) !all_list
      end
    in
    let newsons =
      match inst_type.b_desc with
	BTvar | BTabstract _ | BTbadType -> []
      |	BTarrow(_,dstidx,_,dst) -> 
	  [ (gen_node, status, (dst, get (!dstidx) inst_nt)) :: List.hd sons ]
      |	BTgroup(_, tyl) -> List.map2 (fun (idx,ty) tylist ->
	  (gen_node, status, (ty, get (!idx) inst_nt)) :: tylist) tyl sons
    in
      if status = Res 
      then ((paramcorresp, (inst_nt, gen_node)::rescorresp, newsons), Some inst_type)
      else (((inst_nt, gen_node)::paramcorresp, rescorresp, newsons), Some inst_type)
	) (!s) (call_accu, None)
  in
  Graph.call paramcorresp [] rescorresp [];
  List.iter instcall sons;
  List.iter (fun (_,_,(ty,_)) -> ty.b_mark <- max_int) tylist
  end

let collect_inst gen_node inst_nt accu =
  let accuref = ref accu in
  let rec col inst_nt =
    accuref := (inst_nt, gen_node) :: !accuref;
    Array.iter col (Graph.get_sons inst_nt);
  in
  col inst_nt;
  !accuref

let rec collect instjustseen genjustseen accu instance_type inst_nt generic_type gen_nt =
     match (generic_type.b_link3, instance_type.b_link3) with
        ([ty1'],_) when ty1' == instance_type -> accu (* Back edge *)
      | (_, [ty2']) when ty2' == generic_type -> accu (* Back edge *)
      | (_::_::_, _::_::_) -> Message.error "NewType.collect" 
                 (fun () -> print_string "Wrong assumptions : ";
                            print_sum instance_type;
                            print_string " and ";
                            print_sum generic_type;
                            print_string " do not correspond.")
      | _ -> if instance_type == generic_type then
     begin
	  sons_corresp := (inst_nt, gen_nt) :: !sons_corresp;
	  accu
             (* to avoid adding the type twice in b_link3 *)
     end
     else
     begin
       generic_type.b_link3 <- instance_type :: generic_type.b_link3;
       instance_type.b_link3 <- generic_type :: instance_type.b_link3;
       let accu1 = if (instjustseen == inst_nt) 
	   && (genjustseen == gen_nt) then accu
	   else (inst_nt, gen_nt) :: accu in
       let accu' = 
       begin
       match (generic_type.b_desc, instance_type.b_desc) with
          ((BTvar | BTabstract _ | BTbadType), (BTvar | BTabstract _))
        | (_, BTbadType) | (_, BTAll) -> accu1
	| (BTAll, _) -> 
	    if !status == Param then Graph.mark inst_nt;
	    accu
        | (BTvar, _) | (BTabstract _, _) | (BTbadType, _) ->
	     (* Instantiation. *)
	    typeinst_accu := (gen_nt, !status, (instance_type, inst_nt)) :: !typeinst_accu;
	    if !status == Param 
	    then collect_inst gen_nt inst_nt accu
	    else accu
    (*  | (_, BTvar) | (_, BTabstract _) -> Error ?? *)
        | (BTarrow(_,idxdst1,_,dst1), BTarrow(_,idxdst2,_,dst2)) -> 
	    collect inst_nt gen_nt accu1 dst2 (get (!idxdst2) inst_nt) dst1 (get (!idxdst1) gen_nt)
        | (BTgroup(_, ty1), BTgroup(_, ty2)) ->
            List.fold_left2 (fun accu (idx_inst, ty_inst) (idx_gen, ty_gen) ->
	      collect inst_nt gen_nt accu ty_inst (get (!idx_inst) inst_nt) ty_gen (get (!idx_gen) gen_nt)) accu1 ty2 ty1
        | (_, _) -> Message.error "NewType.collect"
                 (fun () -> print_sum instance_type;
                            print_string " and ";
                            print_sum generic_type;
                            print_string " do not correspond.")
       end in
       generic_type.b_link3 <- List.tl generic_type.b_link3;
       instance_type.b_link3 <- List.tl instance_type.b_link3;
       accu'
   end

let rec collect_param_rec accu inst_type inst_nt gen_type gen_nt =
   match (inst_type.b_desc, inst_nt.e_paramson, gen_type.b_desc, gen_nt.e_paramson) with
     (BTarrow(_,idxdsti,pari,dsti), Some parami, BTarrow(_,idxdstg,parg,dstg), Some paramg) ->
         collect_param_rec 
	 (collect dummynode dummynode accu pari parami parg paramg)
	 dsti (get (!idxdsti) inst_nt) dstg (get (!idxdstg) gen_nt)
   | _ -> iter_params Graph.mark (inst_type, inst_nt); accu
       	   (* If the instance type still contains parameters,
	      they may escape. This can happen for example with
	      'a -> 'a instantiated into (int -> int) -> int -> int
	      the second parameter (int) may escape *)

let call (instance_type, inst_nt) (generic_type, gen_nt) =
  status := Res;
  sons_corresp := [];
  let resultcorresp = 
    collect dummynode dummynode [] instance_type inst_nt generic_type gen_nt in
  let resultsonscorresp = !sons_corresp in
  sons_corresp := [];
  status := Param;
  let paramcorresp = 
    collect_param_rec [] instance_type inst_nt generic_type gen_nt in
  let paramsonscorresp = !sons_corresp in
  sons_corresp := [];
  Graph.call paramcorresp paramsonscorresp resultcorresp resultsonscorresp;
  instcall !typeinst_accu;
  typeinst_accu := []

let call = Profile.f2 "NewType.call" call
