(***********************************************************************)
(*                                                                     *)
(*              Fast Escape Analysis - Environment                     *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1998 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Ttype
open Tsolver

module NewType :
  sig
    type t_type_name = Ttype.t_type_name
    val tuple_name : t_type_name
    type t = newtype_t
    val same_name : t_type_name -> t_type_name -> bool
    val functional : t -> bool
    val arity : t -> int
    val extract_fun : t -> t * t
    val extract_constr_list : t -> Types.constructor_description -> t list
    val set_constr : t -> Types.label_description -> t -> t
    val extract_list : t -> 'a list -> t list
    val extract_list_n : t -> int -> t
    val iter_params : (Graph.t -> unit) -> t -> unit
    val result_mark : t -> int
    val generalize : t -> t -> unit
    val all : t
    val nostack : graph_t
    val copy : int -> t -> t
  end

module CTree :
  sig
    type letone =
      Asttypes.rec_flag * Asttypes.stack_flag * int *
      (Typedtree.pattern * Typedtree.expression) list
    type blocks = lettree list
    and lettree = | LetGroup of blocks | LetOne of letone
    val eval : blocks -> letone * blocks
    val clearone : blocks -> blocks
  end

module VarEnv :
  sig
    type t = NewType.t Ident.tbl
    val create : unit -> 'a Ident.tbl
    val update : Ident.t -> NewType.t -> t -> NewType.t Ident.tbl
    val lookup : Ident.t -> 'a Ident.tbl -> 'a option
  end

type expression =
  { e_desc: expres_desc;
    e_loc: Location.t;
    e_type: Types.type_expr;
    e_env: Env.t;
    e_tail: bool;
    e_tailcalls: bool;
    e_unknown_tailcalls: bool;
    mutable e_type_info: NewType.t;
    mutable e_stackop: Graph.t;
    mutable e_stacktailop: Graph.t }
and expres_desc =
  | Eexp_ident of Path.t * Types.value_description
  | Eexp_constant of Asttypes.constant
  | Eexp_let of Asttypes.rec_flag * Asttypes.stack_flag *
                  (Typedtree.pattern * expression) list * expression
  | Eexp_function of (Typedtree.pattern * expression) list
  | Eexp_apply of expression * expression list
  | Eexp_tuple_apply of expression * expression list
  | Eexp_match of expression * (Typedtree.pattern * expression) list
  | Eexp_try of expression * (Typedtree.pattern * expression) list
  | Eexp_tuple of expression list
  | Eexp_construct of Types.constructor_description * expression list
  | Eexp_record of (Types.label_description * expression) list
  | Eexp_field of expression * Types.label_description
  | Eexp_setfield of expression * Types.label_description * expression
  | Eexp_array of expression list
  | Eexp_ifthenelse of expression * expression * expression option
  | Eexp_sequence of expression * expression
  | Eexp_while of expression * expression
  | Eexp_for of Ident.t * expression * expression * Asttypes.direction_flag *
                  expression
  | Eexp_when of expression * expression
  | Eexp_send of expression * string
  | Eexp_new of Path.t
  | Eexp_instvar of Path.t * Path.t
  | Eexp_setinstvar of Path.t * Path.t * expression
  | Eexp_override of Path.t * (Path.t * expression) list

module IdentInfo :
  sig
    type t = identinfo_t
    val type_expr : t -> NewType.t
    val empty : NewType.t -> bool -> identinfo_t
    val tupled : Typedtree.expression -> bool
    val tupledE : expression -> bool
    val arity_according_exp : Typedtree.expression -> int
    val build : NewType.t -> bool -> bool -> int -> identinfo_t
    val print_info : identinfo_t -> unit
    val show_graph : Ident.t -> identinfo_t -> unit
    val polymorphic_instantiation : identinfo_t -> NewType.t -> unit
    val lookup_primitive : string -> NewType.t -> identinfo_t
    val box_float : string -> bool
    val is_an_allocation : string -> bool
  end

module BuildType :
  sig
    type t = buildtype_t
    val print : t -> unit
    val annotate : buildtype_t_desc -> t
    val annotate2 : int -> buildtype_t_desc -> t
    val copy_type1 : t -> t
    val copy_type : t -> t
    val cleanup : t -> unit
    val declsubstitute1 : t list -> t list -> t -> t
    val declsubstitute : t list -> t list -> t -> t
  end

