(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Format
open Ttype
open Tsolver
open Tprinttyp
open Tlinksons

(**** Instantiation ****)

let get = Graph.get

let rec iternt f nodetree =
  f nodetree;
  Array.iter (iternt f) (Graph.get_sons nodetree)

let rec fast_instres gen_nt inst_nt =
  if inst_nt == gen_nt then
    Message.error "instantiate" (fun () ->
      print_string "Types should never be physically the same when instantiating")
  else
    begin
      Graph.to_replace_by_node gen_nt inst_nt;
      match gen_nt.e_sons with
	Expanded gsons ->
	  let isons = Graph.get_sons inst_nt in
	  for i = 0 to Array.length gsons - 1 do
	    fast_instres gsons.(i) isons.(i)
	  done
      |	Compact cptdesc -> ()
    end

let rec instantiate_res same_scc instance_type inst_nt generic_type gen_nt =
  if instance_type == generic_type then 
    fast_instres gen_nt inst_nt
    (* to avoid adding the type twice in b_link3 *)
  else
    match (generic_type.b_link3, instance_type.b_link3) with
      ([ty1'],_) when ty1' == instance_type -> () (* Back edge *)
    | (_, [ty2']) when ty2' == generic_type -> () (* Back edge *)
    | (_::_::_, _::_::_) -> Message.error "NewType.instantiate_res" 
          (fun () -> 
	    print_string "Wrong assumptions : ";
            print_sum instance_type;
            print_string " and ";
            print_sum generic_type;
            print_string " do not correspond.")
    | _ -> 
	if not same_scc then
	  Graph.to_replace_by_type gen_nt (generic_type, (instance_type, inst_nt));
       match gen_nt.e_sons with
	 Compact cptdesc -> ()
       | Expanded _ -> 
       generic_type.b_link3 <- instance_type :: generic_type.b_link3;
       instance_type.b_link3 <- generic_type :: instance_type.b_link3;
       begin
       match (generic_type.b_desc, instance_type.b_desc) with
          (BTvar, _) | (BTabstract _, _) | (BTbadType, _) -> ()
	| (_, BTAll) | (_, BTbadType) -> 
	    iternt (fun n -> Graph.to_replace_by_node n inst_nt)
	      gen_nt
        | (BTarrow(_,idxdst1,_,dst1), BTarrow(_,idxdst2,_,dst2)) ->
	    instantiate_res ((!idxdst2 == -1) && (!idxdst1 == -1)) dst2 (get (!idxdst2) inst_nt) dst1 (get (!idxdst1) gen_nt)
        | (BTgroup(_, ty1), BTgroup(_, ty2)) ->
             List.iter2 (fun (idx_inst, ty_inst) (idx_gen, ty_gen) ->
	       instantiate_res ((!idx_inst == -1) && (!idx_gen == -1)) ty_inst (get (!idx_inst) inst_nt) ty_gen (get (!idx_gen) gen_nt)
	     ) ty2 ty1
        | (_, _) -> Message.error "NewType.instantiate_res"
                 (fun () -> print_sum instance_type;
                            print_string " and ";
                            print_sum generic_type;
                            print_string " do not correspond.")
       end;
       generic_type.b_link3 <- List.tl generic_type.b_link3;
       instance_type.b_link3 <- List.tl instance_type.b_link3

(* Careful : if we have inst_param, which corresponds to gen_param linked
to gen_res, which corresponds to inst_res, gen_param and gen_res may
be different types in the same scc, so can be inst_param and inst_res.
So they may not correspond according to linksons

Solution : we first look for the corresponding type, thanks to the
corresp function. The corresponding types found are the ones which have a
son numbered 0 (the first son of the scc)

NB : This is only correct because the generic types are directly built
from TranslType.expr, so they are at least as unfolded as the instance
types. Therefore, when looking for the first son of the scc of the generic
type, we are sure that we do not go outside of the scc in the instance type.
This property is checked is goto_pos  *)

exception NoCfc

let rec find_son_pos ty =
  if ty.b_mark = max_int then
    begin
      ty.b_mark <- 0;
      let res = 
      match ty.b_desc with
	BTvar | BTabstract _ | BTbadType | BTAll -> raise NoCfc
      |	BTarrow(_,idxdst,_,dst) -> 
	  if (!idxdst) == 0 then Some([]) else
	  if (!idxdst) == -1 then
	    match find_son_pos dst with
	      None -> None
	    | Some l -> Some(1 :: l)
	  else None
      |	BTgroup(_,l) -> find_list_son 0 l
      in
      ty.b_mark <- max_int;
      res
    end
  else
    None

and find_list_son i = function [] -> None
  | ((idx,ty)::l) -> if !idx == 0 then Some([]) else 
    if !idx == -1 then 
      match find_son_pos ty with
      	None -> find_list_son (i+1) l
      | Some l -> Some(i::l)
    else find_list_son (i+1) l

let rec goto_pos l ty =
  match l with
    [] -> ty
  | (i::l) -> 
  match ty.b_desc with
    BTvar | BTabstract _ | BTbadType | BTAll -> raise NoCfc
  | BTarrow(_,idxdst,_,dst) ->
      if !idxdst != -1 then
      	Message.error "NewType.corresp" (fun () -> print_string
	    "We go outside of the top scc of the instance type");
      goto_pos l dst
  | BTgroup(_,tyl) -> 
      let (idx, ty) = List.nth tyl i in
      if !idx != -1 then
	Message.error "NewType.corresp" (fun () -> print_string
	    "We go outside of the top scc of the instance type");
      goto_pos l ty

let corresp inst_param gen_param gen_res inst_res = 
  if gen_param == gen_res then (inst_param, inst_res) else
  try
  ((match find_son_pos gen_param with
    None -> (* There is no son. The correspondance does not matter,
	       we must only make sure that the types really correspond
	       to avoid any error *) gen_param
  | Some poslist -> goto_pos poslist inst_param),
   (match find_son_pos gen_res with
    None -> (* There is no son. The correspondance does not matter,
	       we must only make sure that the types really correspond
	       to avoid any error *) gen_res
  | Some poslist -> goto_pos poslist inst_res))
  with NoCfc -> (inst_param, inst_res)

let rec fast_instparam idx param_ty gen_nt inst_nt =
  if gen_nt == inst_nt then
    Message.error "instantiate" (fun () -> print_string
	"Types should never be physically the same when instantiating")
  else if param_ty.b_mark = max_int then
    try
      if idx != -1 then Graph.copy_link gen_nt inst_nt;
      match gen_nt.e_sons with
	Expanded gsons ->
	  param_ty.b_mark <- 0;
	  begin
	  match param_ty.b_desc with
           BTvar | BTabstract _ | BTbadType | BTAll -> ()
         | BTarrow(_,idxdst,_,dst) ->
	    fast_instparam (!idxdst) dst (get (!idxdst) gen_nt) (get (!idxdst) inst_nt)
         | BTgroup(_, tyl) ->
             List.iter (fun (idx, ty) ->
	       fast_instparam (!idx) ty (get (!idx) gen_nt) (get (!idx) inst_nt)
	     ) tyl
	  end;
	  param_ty.b_mark <- max_int
      |	Compact cptdesc ->
	  Graph.cpt_copy_link (fun (gen_res, (inst_res_ty, inst_res_nt)) -> 
	    let (param_ty, res_ty) = corresp param_ty param_ty gen_res inst_res_ty in
	    linksons param_ty inst_nt res_ty inst_res_nt) cptdesc inst_nt
    with Graph.EscapeExc ->
      Graph.mark inst_nt

let rec instantiate_param same_scc instance_type inst_nt generic_type gen_nt =
  if instance_type == generic_type then 
    fast_instparam 1 instance_type gen_nt inst_nt
    (* to avoid adding the type twice in b_link3 *)
  else
    match (generic_type.b_link3, instance_type.b_link3) with
      ([ty1'],_) when ty1' == instance_type -> () (* Back edge *)
    | (_, [ty2']) when ty2' == generic_type -> () (* Back edge *)
    | (_::_::_, _::_::_) -> Message.error "NewType.instantiate_param" 
          (fun () -> 
	    print_string "Wrong assumptions : ";
            print_sum instance_type;
            print_string " and ";
            print_sum generic_type;
            print_string " do not correspond.")
    | _ -> 
       try
	 if not same_scc then Graph.copy_link gen_nt inst_nt;
         match (generic_type.b_desc, instance_type.b_desc) with
           (BTvar, _) | (BTabstract _, _) | (BTbadType, _) -> 
             Graph.iter_type_replacers (fun (gen_ty, (res_ty, res_nt)) -> 
	       match gen_ty.b_desc with
		 BTvar | BTabstract _ | BTbadType ->
		  linksons instance_type inst_nt res_ty res_nt
	       | _ ->
		  Graph.linksource inst_nt res_nt) gen_nt inst_nt
    (*   | (BTAll, _) -> Graph.mark inst_nt
      	 Already handled by "try ... with Graph.EscapeExc -> ..." *)
	 | _ ->
	  match gen_nt.e_sons with
	    Compact cptdesc -> 
	      Graph.cpt_copy_link (fun (gen_res, (inst_res_ty, inst_res_nt)) ->
		let (param_ty, res_ty) = corresp instance_type generic_type gen_res inst_res_ty in
		linksons param_ty inst_nt res_ty inst_res_nt) cptdesc inst_nt
	  | Expanded _ ->
         match (generic_type.b_desc, instance_type.b_desc) with
           (_, BTAll) | (_, BTbadType) -> ()
         | (BTarrow(_,idxdst1,_,dst1), BTarrow(_,idxdst2,_,dst2)) ->
	     generic_type.b_link3 <- instance_type :: generic_type.b_link3;
	     instance_type.b_link3 <- generic_type :: instance_type.b_link3;
	     instantiate_param ((!idxdst2 == -1) && (!idxdst2 == -1)) dst2 (get (!idxdst2) inst_nt) dst1 (get (!idxdst1) gen_nt);
	     generic_type.b_link3 <- List.tl generic_type.b_link3;
	     instance_type.b_link3 <- List.tl instance_type.b_link3
         | (BTgroup(_, ty1), BTgroup(_, ty2)) ->
	     generic_type.b_link3 <- instance_type :: generic_type.b_link3;
	     instance_type.b_link3 <- generic_type :: instance_type.b_link3;
             List.iter2 (fun (idx_inst, ty_inst) (idx_gen, ty_gen) ->
	       instantiate_param ((!idx_inst == -1) && (!idx_gen == -1)) ty_inst (get (!idx_inst) inst_nt) ty_gen (get (!idx_gen) gen_nt)
	     ) ty2 ty1;
	     generic_type.b_link3 <- List.tl generic_type.b_link3;
	     instance_type.b_link3 <- List.tl instance_type.b_link3
         | (_, _) -> Message.error "NewType.instantiate_param"
                 (fun () -> print_sum instance_type;
                            print_string " and ";
                            print_sum generic_type;
                            print_string " do not correspond.")
	with Graph.EscapeExc ->
          (* Imperative escapement of the whole parameter *)
	  Graph.mark inst_nt

let rec instantiate_param_rec inst_type inst_nt gen_type gen_nt =
   match (inst_type.b_desc, inst_nt.e_paramson, gen_type.b_desc, gen_nt.e_paramson) with
     (BTarrow(_,dsti_idx,pari,dsti), Some parami, BTarrow(_,dstg_idx,parg,dstg), Some paramg) ->
         instantiate_param false pari parami parg paramg;
         instantiate_param_rec dsti (get (!dsti_idx) inst_nt) dstg (get (!dstg_idx) gen_nt)
   | _ -> iter_params Graph.mark (inst_type, inst_nt)
       	   (* If the instance type still contains parameters,
	      they may escape. This can happen for example with
	      'a -> 'a instantiated into (int -> int) -> int -> int
	      the second parameter (int) may escape *)

let instantiate (instance_type, inst_nt) (generic_type, gen_nt) =
   instantiate_res false instance_type inst_nt generic_type gen_nt;
   instantiate_param_rec instance_type inst_nt generic_type gen_nt;
   Graph.cleanup_replace gen_nt

let instantiate = Profile.f2 "NewType.instantiate" instantiate
