(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

open Format
open Ttype
open Tsolver

let rec compact nt =
  match nt.e_sons with
    Compact cptdesc -> cptdesc.cpt_sons
  | Expanded ntsons -> { d_sons = Array.map (fun n -> compact n) ntsons;
			 d_paramson = None}

let rec mark_all nt =
  let newsons = Compact
	      	{ cpt_pathEdge = [];
		  cpt_linksource = Nset.empty;
		  cpt_linksons = Nset.empty;
		  cpt_rets = [];
		  cpt_starts = [];
		  cpt_sons = compact nt}
  in
  let res = Graph.newnode newsons None in
  Graph.mark res;
  res

let sons_i i nt = match nt.e_sons with
  Compact _ -> raise Not_found
| Expanded ntar -> if Array.length ntar > i then ntar.(i) else raise Not_found

(* prop_linksons = {x \in newnode.norm | sons_i i x \in cptdesc.cpt_linksons}
 *)

let get_list i linksons norm =
  Nset.fold (fun a rest -> 
    if Nset.mem (sons_i i a) linksons then Nset.add a rest else rest) 
    norm Nset.empty

let find_links norm ntsons =
  let rec find i =
  if i >= Array.length ntsons then raise Not_found;
  match ntsons.(i).e_sons with
    Compact cptdesc ->
      (cptdesc.cpt_linksource, get_list i cptdesc.cpt_linksons norm)
  | Expanded ntar ->
      if ntar = [| |] then find (i+1) else raise Not_found
  in
  try
    find 0
  with Not_found ->
    (Nset.inter ntsons.(0).norm norm, Nset.diff norm ntsons.(0).norm)

let check linksource linksons norm ntsons =
  if (not (Nset.equal (Nset.inter linksons linksource) Nset.empty))
  || (not (Nset.equal (Nset.union linksons linksource) norm)) 
  then raise Not_found;
  for i = 0 to Array.length ntsons - 1 do
    let linksons_i = Nset.fold (fun n accu -> Nset.add (sons_i i n) accu)
	linksons Nset.empty in
    if (not (Nset.equal (Nset.union linksons_i linksource) ntsons.(i).norm))
       || (ntsons.(i).node_replacer == Escape) then raise Not_found;
    match ntsons.(i).e_sons with
      Compact cptdesc ->
	if (not (Nset.equal cptdesc.cpt_linksource linksource))
           || (not (Nset.equal cptdesc.cpt_linksons linksons_i))
	then raise Not_found
    | Expanded ntar -> if ntar <> [| |] then raise Not_found
  done

let rec reduce_param nt =
      try
	(* match nt.node_replacer with
	   OneNode newnode -> newnode
         | _ -> *)
	let ntrep = Graph.get_node_replacers nt in
	let newsons = match nt.e_sons with
	  Expanded ntsons -> Expanded (Array.map reduce_param ntsons)
	| Compact cptdesc -> 
	    try
	      Compact
	      	{ cpt_pathEdge = [];
		  cpt_linksource = Graph.cpt_get_source_node_replacers nt;
		  cpt_linksons = Graph.cpt_get_node_replacers nt;
		  cpt_rets = [];
		  cpt_starts = [];
		  cpt_sons = cptdesc.cpt_sons}
	    with Graph.ExpandedMet ->
	      Expanded (Array.map reduce_param (Graph.get_sons nt))
	in
	let newnode = Graph.newnode newsons None in
	newnode.norm <- ntrep;
	begin
	  match newsons with
	    Compact _ -> ()
	  | Expanded ntsons -> if ntsons <> [| |] then
              (* Try to compact *)
	      let (linksource, linksons) = find_links newnode.norm ntsons in
	      try
		check linksource linksons newnode.norm ntsons;
		newnode.e_sons <- Compact
		     { cpt_pathEdge = [];
		       cpt_linksource = linksource;
		       cpt_linksons = linksons;
		       cpt_rets = [];
		       cpt_starts = [];
		       cpt_sons = { d_sons = Array.map (fun n -> 
			 match n.e_sons with
			   Compact cptdesc -> cptdesc.cpt_sons
			 | Expanded ntar -> if ntar = [| |] then
			     { d_sons = [| |]; d_paramson = None}
			   else raise Not_found) ntsons;
				    d_paramson = None }
		     }
	      with Not_found -> ()
	end;
	(* nt.node_replacer <- OneNode newnode; *)
	newnode
      with Graph.EscapeExc -> 
        (* Imperative escapement of the whole parameter *)
        mark_all nt

let rec reduce_res nt =
  match nt.node_replacer with
    NoRep | Escape ->
        let newsons = match nt.e_sons with
	  Expanded ntsons -> Expanded (Array.map reduce_res ntsons)
	| Compact cptdesc -> Compact 
	      { cpt_pathEdge = [];
		cpt_linksource = Nset.empty;
		cpt_linksons = Nset.empty;
		cpt_rets = [];
		cpt_starts = [];
		cpt_sons = cptdesc.cpt_sons}
 	in
	let newparamson = 
	  match nt.e_paramson with
            None -> None
          | Some param ->
              Graph.incr_mark();
              Some (reduce_param param)
	in
	let newnode = Graph.newnode newsons newparamson in	      
	if nt.node_replacer = Escape then
	  newnode.node_replacer <- Escape;
	nt.node_replacer <- OneNode newnode;
	newnode
(*| OneNode newnode -> newnode *)
  | _ -> Message.error "reduce" 
	(fun () -> print_string "Node met twice")


let rec mark_linked_nodes nt =
  Nset.iter (fun n -> n.replacer_visited <- true) nt.norm;
  begin
  match nt.e_paramson with
    None -> ()
  | Some n -> mark_linked_nodes n
  end;
  match nt.e_sons with
    Compact cptdesc ->
      Nset.iter (fun n -> n.replacer_visited <- true) cptdesc.cpt_linksource;
      Nset.iter (fun n -> n.replacer_visited <- true) cptdesc.cpt_linksons;
  | Expanded ntar ->
      for i = 0 to Array.length ntar - 1 do
	mark_linked_nodes ntar.(i)
      done

let rec compress_unlinked_nodes nt =
  match nt.e_sons with
    Compact cptdesc -> ()
  | Expanded ntar -> if ntar <> [| |] then
      begin
      for i = 0 to Array.length ntar - 1 do
	compress_unlinked_nodes ntar.(i);
      done;
      try
      for i = 0 to Array.length ntar - 1 do
	if ntar.(i).replacer_visited then raise Not_found;
	match ntar.(i).e_sons with
	  Compact _ -> ()
	| Expanded ar -> if ar <> [| |] then raise Not_found
      done;
      if (nt.e_paramson = None) then
	nt.e_sons <- Compact
	      	{ cpt_pathEdge = [];
		  cpt_linksource = Nset.empty;
		  cpt_linksons = Nset.empty;
		  cpt_rets = [];
		  cpt_starts = [];
		  cpt_sons = compact nt}
      with Not_found -> ()
      end

let rec cleanup_repvisit nt =
  nt.replacer_visited <- false;
  match nt.e_sons with
    Compact cptdesc -> ()
  | Expanded ntar ->
      Array.iter cleanup_repvisit ntar

let rec cleanup_replace n =
  begin
  match n.node_replacer with
    OneNode m ->
      n.node_replacer <- m.node_replacer;
      m.node_replacer <- NoRep
  | _ -> ()
  end;
  n.type_replacer <- [];
  match n.e_sons with
    Expanded nsons -> Array.iter cleanup_replace nsons
  | Compact cptdesc -> ()

let reduce (ty, nt) =
  let nt' = reduce_res nt in
  (*iter_params Graph.cleanup_replace nt;*)
  cleanup_replace nt;
  mark_linked_nodes nt';
  compress_unlinked_nodes nt';
  cleanup_repvisit nt';
  (ty, nt')

let reduce = Profile.f1 "NewType.reduce" reduce
