(***********************************************************************)
(*                                                                     *)
(*                         Fast Escape Analysis                        *)
(*                                                                     *)
(*                  Bruno Blanchet, ENS Ulm - INRIA Rocquencourt       *)
(*                                                                     *)
(*  Copyright 1995 Institut National de Recherche en Informatique et   *)
(*  Automatique.  Distributed only by permission.                      *)
(*                                                                     *)
(***********************************************************************)

(* $Id: ttrtype.ml,v 1.32 2001/02/21 10:14:24 bblanche Exp $ *)

open Types
open Typedtree
open Format
open Tsolver
open Tenv
open Tenvir
open Ttype

type repr_t =
    RTvar
  | RTtuple of repr_t list
  | RTarrow of repr_t * repr_t
  | RTconstr of Path.t * repr_t list
  | RTobject

module TypeMap = Map.Make (struct
                              type t = repr_t

                              let compare (x :t) (y :t) = compare x y
                           end)

module TranslType =
struct

open NewType
open BuildType

(* Values of b_mark :
   Just after creating the expression = -1
   During number1  = age
   At the end of number1 = max_int-1
   Number2 passes from max_int-1 to max_int
*)

let node_list = ref ([] : Tenv.BuildType.t list)

let age = ref 0

let rec number1 ty = 
   (* b_mark = age of the node *)
   if ty.b_mark = -1 then
     begin
         incr age;
         node_list := ty :: !node_list;
         ty.b_mark <- !age;
         let age_sons = 
           match ty.b_desc with
	     BTvar | BTbadType | BTAll | BTabstract _ -> !age
           | BTarrow(paridx,dstidx,par,dst) -> 
               let age_par =
		 let (age_par, same_scc) = number1 par in
		 if same_scc then paridx := -3 else paridx := -4;
		 age_par
               and age_dst =
		 let (age_dst, same_scc) = number1 dst in
		 if same_scc then dstidx := -3 else dstidx := -4;
		 age_dst
               in 
               min age_par age_dst
           | BTgroup(d,content) -> 
               number_type_list content
         in
	 if age_sons >= ty.b_mark then
                      (* Entry node of a scc *)
	   let rec mark_scc = function [] -> []
	     | (a::l) -> a.b_mark <- max_int-1;
		 if a == ty then l else mark_scc l
	   in
	   node_list := mark_scc !node_list;
	   (age_sons, false)
         else 
           (age_sons, true)
     end
   else (ty.b_mark, ty.b_mark < max_int-1)

and number_type_list = 
      function [] -> max_int
             | ((idx,a)::l) -> 
                 let age_l = number_type_list l
                 and age_a =
		   let (age_a, same_scc) = number1 a in
		   if same_scc then idx := -3 else idx := -4;
		   age_a
                 in min age_l age_a

(* Sets the number of sons for all nodes of the scc to n *)

let rec affect_nsons n ty =
  if (ty.n_sons = -1) then
    begin
      ty.n_sons <- n;
      match ty.b_desc with
	BTvar | BTbadType | BTAll | BTabstract _ -> ()
      |	BTarrow(_,dstidx,_,dst) -> if !dstidx = -1 then affect_nsons n dst
      |	BTgroup(d,content) -> List.iter (fun (idx,ty) ->
	  if !idx = -1 then affect_nsons n ty) content
    end

let rec number2 cntr ty = 
   if ty.b_mark = max_int-1 then
     begin
       ty.b_mark <- max_int;
       match ty.b_desc with
	     BTvar | BTbadType | BTAll | BTabstract _ -> cntr
           | BTarrow(paridx,dstidx,par,dst) ->
	       (* We ignore the parameters when they are in the same scc *)
	       if (!paridx = -3) || (cntr != 0) || (!dstidx = -3) 
	       then paridx := -1 
	       else
		 begin
		   paridx := 1;
		   if par.n_sons = -1 then affect_nsons (number2 0 par) par
		 end;
	       if !dstidx = -4 then
		 begin
		   if dst.n_sons = -1 then affect_nsons (number2 0 dst) dst;
		   dstidx := cntr;
		   cntr + 1
		 end
	       else
		 begin
		   dstidx := -1;
		   number2 cntr dst
		 end
           | BTgroup(d,content) -> 
               number_type_list cntr content
     end
   else cntr

and number_type_list cntr = 
  function [] -> cntr
         | ((idx,a)::l) -> number_type_list 
                 (if !idx = -4 then
		   begin
		     if a.n_sons = -1 then affect_nsons (number2 0 a) a;
		     idx := cntr;
		     cntr + 1
		   end
		 else
		   begin
		     idx := -1;
		     number2 cntr a
		   end) l 

(* Phase 1 : Traduction des declarations de types dans un langage plus
   commode pour les traitements *)

exception Baddep

let modname = ref ""

let id_object = Ident.create "object"
let decl_object () = 
  let name = {d_modname = !modname; d_name = id_object} in
  let decl = 
    { e_name = name; e_type_params = []; e_color = 4;
      e_orig_decl = NoDecl; e_type_kind = annotate BTbadType}
  in
  decl.e_type_kind.b_desc <- BTabstract(name, decl, []);
  decl.e_type_kind

let distinct args = 
    List.iter (fun a -> a.b_mark <- -2) args;
    List.iter (fun a -> a.b_mark <- a.b_mark + 1) args;
    List.for_all (fun a -> (a.b_mark = -1) || (a.b_mark <- -1;false)) args

let decl env decllist =
   let instsubst = ref [] in
   let (env, decllist') = List.fold_right (fun (name, info) (env, decllist') ->
      let result = {e_name = {d_modname = !modname; d_name = name};
                    e_type_params = [];
                    e_type_kind = annotate BTbadType;
                    e_orig_decl = OldDecl info;
                    e_color = 0}
      in (Envir.add_type name result env, result :: decllist')) decllist (env, [])
   in

   let rec expr ty =
      let ty = Ctype.repr ty in
          match ty.desc with
               Tvar -> (try 
                     List.assq ty !instsubst 
                  with Not_found ->
                     Message.error "TranslType.decl.expr" 
                     (fun () -> print_string "Type variable not found"))
             | Tarrow(ty1,ty2) -> 
                  annotate (BTarrow(ref (-2), ref (-2), expr ty1, expr ty2))
             | Ttuple(types) -> 
                  annotate (BTgroup(tuple_name, List.map (fun ty -> (ref (-2), expr ty)) types))
             | Tconstr(name, args, _) -> 
                  let decl = Envir.lookup_type name env
                  and args = List.map expr args in
  (* The real test we want is :
	if decl is in the current scc, it must be called with distinct
variables as parameters, these variables will be the parameters of decl,
and the translation is decl.e_type_kind
	else, the translation is 
declsubstitute decl.e_type_params args decl.e_type_kind
     But we do not know whether decl is in the current scc or not, so
we change this test into :
	if the translation of decl has not begun (e_color = 0),
	  - if args are distinct variables, decl may be in the current scc
            translate decl with args as parameters, and the translation of
            the call to decl is decl.e_type_kind
          - if args are not distinct variables, decl must not be in the
            current scc, translate decl with other parameters, and
	    substitute to translate the call.
	if the translation of decl has begun (e_color = 1),
	  - if args are equal to the parameters of decl, it is a correct
	    back edge, the translation is decl.e_type_kind
	  - else, this is a wrong back edge, raise Baddep
	if the translation of decl is finished (e_color = 4),
	  - if args are equal to the parameters of decl, it is a correct
	    forward edge, decl may be or not in the current scc, anyway
            translate the call by decl.e_type_kind
	  - else, decl MUST not be in the same scc, but we do not still
	    know if it is (there may be bad dependencies). The translation
	    is : declsubstitute decl.e_type_params args decl.e_type_kind.
	    and the bad dependencies will appear if a not completely
            translated type is met (this type will still be BTbadType)
*)
		  if decl.e_color > 0 then
                  begin
                      if List.for_all2 (==) args decl.e_type_params then
                         decl.e_type_kind
                      else if decl.e_color = 1 then
                         raise Baddep
                      else (* decl.e_color = 4 *)
		         declsubstitute1 decl.e_type_params args decl.e_type_kind
                  end
                  else (* decl.e_color = 0 *)
                  begin
                      if (List.for_all (fun v -> v.b_desc = BTvar) args)
                      && (distinct args) then
                      begin
                         declar (Some args) decl;
                         decl.e_type_kind
                      end
                      else
                      begin
                         declar None decl;
                         declsubstitute1 decl.e_type_params args decl.e_type_kind
                      end
                  end
             | Tnil | Tfield _ ->
                  Message.error "Transltype.decl" (fun () -> print_string
                   "Object fields cannot be in place of a type")
(* TO DO *)  | Tobject _ -> decl_object()
             | Tvar | Tlink _ ->
                  Message.error "Transltype.decl" (fun () -> print_string
                   "Tlink and Tvar should have been already treated")


   and declar fixvars decl = 
     let orig_decl = match decl.e_orig_decl with
         OldDecl x -> x
       | _ -> Message.error "declar" (fun () -> print_string 
                   "declaration empty")
     in
     try
     begin
        if List.exists (fun v -> (Ctype.repr v).desc <> Tvar)
           orig_decl.type_params then
        begin
           (match fixvars with
              None -> decl.e_type_params <- List.map (fun v -> annotate
                        BTvar) orig_decl.type_params
            | Some args -> decl.e_type_params <- args);
           Message.debug 0 (fun () -> print_string "Formal parameter <> Tvar");
           raise Baddep
        end
        else
        match fixvars with
           None ->
               decl.e_type_params <- List.map (fun v -> let x = annotate BTvar 
                in instsubst := (Ctype.repr v,x) :: !instsubst; x)
                orig_decl.type_params
         | Some args ->
               decl.e_type_params <- args;
               List.iter2 (fun v a -> 
                  instsubst := (Ctype.repr v,a) :: !instsubst)
                  orig_decl.type_params args               
     end;
     begin
     match orig_decl.type_manifest with
       Some e -> decl.e_type_kind <- expr e
                 (* An abbreviation is translated each time it is met =>
                    does not work for recursive abbreviations !!
                    This is ok for Ocaml 1.06 because only recursive
                    abbreviations that go through objects are accepted,
                    and here I ignore objects. This does not work in general
                    for Ocaml 1.05 *)
     | None -> 
         decl.e_type_kind <- annotate BTbadType;
         decl.e_color <- 1;
         match orig_decl.type_kind with
           Type_abstract -> 
	     decl.e_type_kind.b_desc <- 
		BTabstract (decl.e_name, decl, decl.e_type_params)
         | Type_variant(varlist) -> 
	          let rec make_variants = function
		    [] -> []
		  | ((_, [])::l) -> make_variants l 
                                    (* Constant constructor ignored *)
		  | ((_, [x])::l) -> (ref (-2),expr x) :: make_variants l
		                     (* One parameter constructor *)
		  | ((_, explist)::l) ->
		      (ref (-2), annotate (BTgroup(tuple_name, 
		       List.map (fun ty -> (ref (-2), expr ty)) explist)))
                       :: make_variants l
			(* Any other constructor *)
		  in
                  decl.e_type_kind.b_desc <-
                   (BTgroup(decl.e_name, make_variants varlist))
         | Type_record(fieldlist) -> 
             decl.e_type_kind.b_desc <- (BTgroup(decl.e_name, List.map 
		(fun (_,_,exp) -> (ref (-2),expr exp)) fieldlist))
     end;
     decl.e_color <- 4;
     decl.e_orig_decl <- NoDecl
     with Baddep ->
       decl.e_type_kind <- annotate BTbadType;
       Message.debug 0 (fun () ->
                     print_string "The type ";
                     Printtyp.ident decl.e_name.d_name;
                     print_string " has bad dependances.");
       decl.e_color <- 4;
       decl.e_orig_decl <- NoDecl;
       raise Baddep

   in List.iter (fun decl -> if decl.e_color = 0 then
           try
              declar None decl
           with Baddep -> ()) decllist';
      List.iter (fun decl -> if decl.e_type_kind.n_sons = -1 then
	begin
	  number1 decl.e_type_kind; 
	  affect_nsons (number2 0 decl.e_type_kind) decl.e_type_kind
	end) decllist';
      env

(* Instantiation of type declarations when applying a functor *)

let rec matchtypes randsig randmod = 
    match (randsig, randmod) with
        (Module mrandsig, Module mrandmod) ->
           Ident.iter (fun typename typesig -> 
             typesig.e_orig_decl <- NewDecl (Ident.find_name (Ident.name
                 typename) mrandmod.types)) mrandsig.types;
           Ident.iter (fun modname modsig ->
             matchtypes modsig (Ident.find_name (Ident.name modname)
                 mrandmod.submod)) mrandsig.submod;
           Ident.iter (fun modname modsig ->
             matchtypes modsig (Ident.find_name (Ident.name modname)
                 mrandmod.submod_types)) mrandsig.submod_types
      | (Functor(sig1, sig2), Functor(mod1, mod2)) ->
           matchtypes sig1 mod1;
           matchtypes sig2 mod2
      | (_, _) -> Message.error "matchtypes" (fun () -> print_string
           "Argument of functor does not correspond to signature")

let ident_fold fn accuanc accunouv tbl =
   let x = ref [] in
   Ident.iter (fun name id -> x := (name, id) :: (!x)) tbl;
   List.fold_left (fun (accu, accuanc, accunouv) (name, id) -> 
     let (id', accuanc', accunouv') = fn accuanc accunouv id in
     (Ident.add name id' accu, accuanc', accunouv'))
     (Ident.empty, accuanc, accunouv) (!x)

let copy_types1 accuanc accunouv typetbl =
   let typelist = ref [] in
   Ident.iter (fun name id -> typelist := id :: (!typelist)) typetbl;
   let (env, accunouv') = List.fold_left (fun (env, decllist') info ->
      match info.e_orig_decl with
        NewDecl result -> 
          (Ident.add result.e_name.d_name result env, result :: decllist')
      | NoDecl ->
      let result = 
	{ e_name = { d_modname = info.e_name.d_modname;
		     d_name = Ident.create (Ident.name info.e_name.d_name)};
          e_type_params = info.e_type_params;
          e_type_kind = annotate BTbadType;
          e_color = 0;
	  e_orig_decl = NoDecl}
      in info.e_orig_decl <- NewDecl result;
         (Ident.add result.e_name.d_name result env, result :: decllist')
      |	_ -> Message.error "TranslType.instantiate" (fun () -> print_string
	    "e_orig_decl not cleaned up"))
     (Ident.empty, accunouv) (!typelist) in
   (env, (List.rev (!typelist)) @ accuanc, accunouv')

let translate accuanc accunouv =
   let rec expr ty =
    match ty.b_link with
       BTlink ty' -> ty' (* Forward  or back edge *)
     | _ -> 
	 match ty.b_desc with
	   BTvar | BTbadType | BTAll -> ty
	 | BTabstract(d_name, decl, params) ->
	     let ty' = 
	       match decl.e_orig_decl with
		 NoDecl -> ty
	       | NewDecl decl' -> declsubstitute1 decl'.e_type_params 
		     (List.map expr params) decl'.e_type_kind 
	       | _ -> Message.error "TranslType.instantiate" (fun () -> 
		   print_string "e_orig_decl not cleaned up")
	     in
	     ty.b_link <- BTlink ty';
	     ty'
	 | BTarrow(_,_,ty1,ty2) -> 
	     let ty' = annotate BTbadType in
             ty.b_link <- BTlink ty';
	     ty'.b_desc <- (BTarrow(ref (-2), ref(-2), expr ty1, expr ty2));
	     ty'
	 | BTgroup(d_name,types) -> 
	     let ty' = annotate BTbadType in
             ty.b_link <- BTlink ty';
	     ty'.b_desc <- (BTgroup(d_name, List.map 
		   (fun (_, ty) -> (ref(-2), expr ty)) types));
	     ty'
   in
   List.iter2 (fun ancdecl nouvdecl ->
     match nouvdecl.e_type_kind.b_desc with
       BTbadType -> 
         nouvdecl.e_type_kind <- expr ancdecl.e_type_kind
     | _ -> () (* This type is in fact part of the parameter of the functor.
		  Do not modify it ! *) ) accuanc accunouv;
   List.iter (fun decl -> 
     BuildType.cleanup decl.e_type_kind;
     decl.e_orig_decl <- NoDecl) accuanc;
   List.iter (fun decl -> 
     decl.e_color <- 4;
     if decl.e_type_kind.n_sons = -1 then
	begin
	  number1 decl.e_type_kind; 
	  affect_nsons (number2 0 decl.e_type_kind) decl.e_type_kind
	end) accunouv

let rec copy_types accuanc accunouv = function
    Module m -> 
       let (types', accuanc', accunouv') = 
          copy_types1 accuanc accunouv m.types in
       let (submod', accuanc'', accunouv'') = 
          ident_fold copy_types accuanc' accunouv' m.submod in
       let (submod_types', accuanc''', accunouv''') =
          ident_fold copy_types accuanc'' accunouv'' m.submod_types in
       (Module {env = m.env; types = types'; submod = submod';
                submod_types = submod_types'}, accuanc''', accunouv''')
  | Functor(sig1, sig2) ->
      (* If we have a functor functor(R1 : S1) -> functor(R2 : S2) -> ...
	 S2 may refer to types in S1. Therefore, when applying the
	 functor to its first parameter M1, the types in S1 must be 
	 substituted in S2 by types in M1. Hence copy_types
	 must also by applied to the parameter S2 of the functor 
	 (here S2 <-> sig1) *)
       let (sig1', accuanc', accunouv') = copy_types accuanc accunouv sig1 in
       let (sig2', accuanc'', accunouv'') = copy_types accuanc' accunouv' sig2
       in (Functor(sig1', sig2'), accuanc'', accunouv'')
       (* let (sig2', accuanc', accunouv') = copy_types accuanc accunouv sig2
       in (Functor(sig1, sig2'), accuanc', accunouv') *)

let instantiate rand_var randenv body =
    matchtypes rand_var randenv;
    let (body', accuanc, accunouv) = copy_types [] [] body in
    translate accuanc accunouv;
    body'

let copy_sig body =
    let (body', accuanc, accunouv) = copy_types [] [] body in
    translate accuanc accunouv;
    body'

(* Translate type expressions to the simpler language *)

let annotate_dum sons paramson =
  {d_sons = sons; d_paramson = paramson}
let dummynt = annotate_dum [| |] None

let rec affecttytont ntarray ty =
  if ty.b_mark = 0 then () else
  begin
    ty.b_mark <- 0;
    begin
      match ty.b_desc with
      	BTvar | BTabstract _ | BTbadType | BTAll ->
      	  Message.error "tytont" (fun () -> print_string
	      "should be handled by tytont")
      | BTarrow(_, dstidx, _, dst) ->
      	  if !dstidx = -1 then 
	    affecttytont ntarray dst
      	  else
	    ntarray.(!dstidx) <- tytont dst
      | BTgroup(_, tylist) ->
      	  List.iter (fun (idx, ty) ->
	    if !idx = -1 then
	      affecttytont ntarray ty
	    else
	      ntarray.(!idx) <- tytont ty) tylist
    end;
    ty.b_mark <- max_int
  end


and tytont ty = 
  match ty.b_link with
    NTlink nt -> nt
  | NoLink ->
  ty.b_mark <- 0;
(* 
      let newsons = Array.create ty.n_sons dummynt in
      let paramson = 
         match ty.b_desc with
           BTarrow(paridx, _, par, _) when !paridx = -1 -> Some (tytont par)
         | _ -> None
      in
      affecttytont newsons ty;
      annotate_dum newsons paramson
*)
  let res =
  match ty.b_desc with
    BTvar | BTabstract _ | BTbadType | BTAll ->
      annotate_dum [| |] None
  | BTarrow(paridx, dstidx, par, dst) ->
      let paramson = if !paridx = -1 then None else Some (tytont par) in
      if !dstidx = -1 then
      	let newsons = Array.create ty.n_sons dummynt in
	affecttytont newsons dst;
	annotate_dum newsons paramson
      else
 	annotate_dum [| tytont dst |] paramson
  | BTgroup(_, tylist) ->
      let newsons = Array.create ty.n_sons dummynt in
      List.iter (fun (idx, ty) ->
	if !idx = -1 then
	  affecttytont newsons ty
	else
	  newsons.(!idx) <- tytont ty) tylist;
      annotate_dum newsons None
  in
  ty.b_mark <- max_int;
  ty.b_link <- NTlink res;
  res

type typemap_t = Ttype.buildtype_t TypeMap.t

let typemap = ref (TypeMap.empty : typemap_t)
  (* This does not memorize submodules definitions : when we close a
     module, we forget all the definitions in this module because the
     names of the declarations change when we change the current module *)

(* Using this function instead of Ctype.repr avoids the segmentation
   fault when compiling typing/mtype.ml. Why ? Perhaps a recursive 
   abbreviation in the abbrev_memo of a TConstr, which would cause
   an infinite loop in the comparison ?

   It should speed up the work of TranslType.expr, by enabling TypeMap.find
   to succeed more often (types which differed only by the presence of Tlink
   or by their level are now identical). 
*)

let rec repr ty = 
   match ty.desc with
     Tlink t -> repr t
   | Tvar -> RTvar
   | Ttuple l -> RTtuple (List.map repr l)
   | Tarrow(p,d) -> RTarrow (repr p, repr d)
   | Tconstr(p, params, _) -> RTconstr(p, List.map repr params)
   | Tobject _ -> RTobject (* Do not treat object types *)
   | Tfield _ | Tnil -> Message.error "TranslType.repr" (fun () -> ())
       (* There should not be object fields outside an object type *)

let rec copynt nt =
    match nt.d_paramson with 
      None -> Graph.newnode (
	if nt.d_sons = [| |] 
      	then Expanded [| |] 
      	else Compact ({ cpt_pathEdge = [];
		      	cpt_linksource = Nset.empty;
		      	cpt_linksons = Nset.empty;
		      	cpt_rets = [];
		      	cpt_starts = [];
		      	cpt_sons = nt})) None
    | Some param ->
	let sons = Expanded (Array.map copynt nt.d_sons) in
	Graph.incr_mark();
	let paramson = Some (copynt param) in
    	Graph.newnode sons paramson

let expr env ty =
  let rec expr' ty =
   try
      TypeMap.find ty !typemap
   with Not_found ->
      let result =
      	match ty with
          RTvar -> annotate2 0 BTvar
        | RTarrow(par,dst) ->
            annotate2 1 (BTarrow(ref 1, ref 0, expr' par, expr' dst))
        | RTtuple(types) -> 
	    let long = List.length types in
	    let n = ref (-1) in
	    let sonsty = List.map (fun ty -> 
	      incr n; 
	      (ref (!n), expr' ty)) types in
            annotate2 long (BTgroup(tuple_name, sonsty))
        | RTconstr(name, args) -> 
            let decl = Envir.lookup_type name env in
	    if args = [] then
	      decl.e_type_kind
	    else
	      declsubstitute decl.e_type_params (List.map expr' args) 
		decl.e_type_kind
 	| RTobject -> decl_object()
      in 
      typemap := TypeMap.add ty result !typemap;
      result
  in
  let res = expr' (repr ty) in
  (res, copynt (tytont res))

let expr = Profile.f2 "TranslType.expr" expr

end

